import pandas as pd
from pandas import Timestamp

from tabstar.preprocessing.dates import series_to_dt


def test_unix_timestamps():
    values = [1500000000, 999999999999990021]
    expected = ['1970-01-01 00:00:01.500000', '2001-09-09 01:46:39.999990021']
    _assert_expected(values, expected)

def test_day_only():
    values = ["2017-10-02"]
    expected = ["2017-10-02 00:00:00"]
    _assert_expected(values, expected)

def test_iso_without_timezone():
    values = ["2022-03-15T12:30:00"]
    expected = ["2022-03-15 12:30:00"]
    _assert_expected(values, expected)

def test_iso_with_timezone():
    values = ["2017-10-02 17:11:16+00"]
    expected = ["2017-10-02 17:11:16"]  # timezone info removed
    _assert_expected(values, expected)

def test_with_bad_quotes():
    values = ["8/3/2017", "4/6/2014"]
    expected = ["2017-08-03 00:00:00", "2014-04-06 00:00:00"]
    _assert_expected(values, expected)

def test_invalid_date():
    values = ["not a date", None]
    expected = [pd.NaT, pd.NaT]
    _assert_expected(values, expected)


def _assert_expected(values, expected):
    """
    Converts the input values using series_to_dt and compares each result
    with the expected values. Expected values can be either a datetime string
    (which is converted to a Timestamp) or pd.NaT for invalid dates.
    """
    results = list(series_to_dt(pd.Series(values)))
    for i, (res, exp) in enumerate(zip(results, expected)):
        if pd.isna(exp):
            assert pd.isna(res), f"Row {i}: expected NaT, got {res}"
        else:
            # Convert expected value to a Timestamp and compare
            assert res == Timestamp(exp), f"Row {i}: expected {Timestamp(exp)}, got {res}"



import numpy as np
import pandas as pd
from torch.utils.data import DataLoader

from torch.utils.data import Dataset

from tabstar.tabstar_verbalizer import TabSTARData
from tabstar.training.hyperparams import GLOBAL_BATCH


class TabSTARDataset(Dataset):
    def __init__(self, data: TabSTARData):
        self.x_txt = data.x_txt
        self.x_num = data.x_num
        if isinstance(data.y, pd.Series):
            self.y = data.y.reset_index(drop=True)
        elif isinstance(data.y, np.ndarray):
            self.y = pd.Series(data.y, dtype=np.float32)
        else:
            # Dummy target for convenience
            self.y = pd.Series(np.zeros(len(data.x_txt)), dtype=np.float32)
        self.d_output = data.d_output

    def __len__(self):
        return len(self.y)

    def __getitem__(self, idx: int):
        x_txt = self.x_txt[idx]
        x_num = self.x_num[idx]
        y = self.y.iloc[idx]
        return x_txt, x_num, y, self.d_output
    
def get_dataloader(data: TabSTARData, is_train: bool, batch_size: int = GLOBAL_BATCH) -> DataLoader:
    dataset = TabSTARDataset(data)
    return DataLoader(dataset, shuffle=is_train, batch_size=batch_size, num_workers=0, collate_fn=collate_fn)


def collate_fn(batch) -> TabSTARData:
    x_txt_batch, x_num_batch, y_batch, d_output_batch = zip(*batch)
    # Assuming all batches have the same d_output, which is correct for finetune
    d_output = d_output_batch[0]
    x_txt = np.stack(x_txt_batch)
    x_num = np.stack(x_num_batch)
    y = pd.Series(y_batch)
    data = TabSTARData(d_output=d_output, x_txt=x_txt, x_num=x_num, y=y)
    return data


from typing import Any

from pandas import Series

from tabstar.preprocessing.nulls import get_valid_values

MAX_NUMERIC_FOR_CATEGORICAL = 50


def is_mostly_numerical(s: Series) -> bool:
    values = get_valid_values(s)
    unique = set(values)
    n_unique = len(unique)
    if n_unique <= MAX_NUMERIC_FOR_CATEGORICAL:
        return False
    non_numerical_unique = [v for v in unique if not is_numeric(v)]
    if len(non_numerical_unique) > 1:
        return False
    return True



def is_numeric(f: Any) -> bool:
    if f is None:
        return False
    if isinstance(f, str):
        return f.isdigit()
    if isinstance(f, (int, float,)):
        return True
    try:
        f = float(f)
        return True
    except ValueError:
        print(f"ValueError: {f} from type {f} cannot be converted to float")
        return False


import datetime
from os import makedirs
from os.path import join, basename
from typing import List, Dict, Callable, Optional

import numpy as np
import torch
from torch import nn
from torch.utils.data import DataLoader


class CheckpointManager:

    def __init__(self, do_average: bool, output_dir: Optional[str]):
        output_dir = output_dir or ".tabstar_checkpoint/"
        self.save_dir = join(output_dir, datetime.datetime.now().strftime("%Y%m%d_%H%M%S"))
        makedirs(self.save_dir, exist_ok=True)
        self.avg_dir = join(self.save_dir, "averaged_model")
        self.best_dir = join(self.save_dir, "best_model")
        self.to_load_dir = self.best_dir
        self.cp_paths: List[str] = []
        self.val_losses: List[float] = []
        self.do_average = do_average
        self.avg_metric: Optional[float] = None

    def save_checkpoint(self, model: nn.Module, epoch: int, val_loss: float):
        """Save a checkpoint for later averaging (only model weights needed)"""
        if not self.do_average:
            return
        checkpoint_path = join(self.save_dir, f"checkpoint_epoch_{epoch}.pt")
        checkpoint = {'model_dict': model.state_dict()}
        torch.save(checkpoint, checkpoint_path)
        self.cp_paths.append(checkpoint_path)
        self.val_losses.append(val_loss)
        assert len(self.cp_paths) == len(self.val_losses) == epoch

    def average_checkpoints(self, model: nn.Module, evaluator: Callable, val_loader: DataLoader):
        if not self.do_average:
            return
        if len(self.cp_paths) < 2:
            print(f"⚠️ Only {len(self.cp_paths)} checkpoint(s) available, skipping averaging")
            return
        best_loss = min(self.val_losses)
        best_idx = int(np.argmin(self.val_losses))
        best_epoch = best_idx + 1
        print(f"🏆 Best checkpoint: Epoch {best_epoch} with loss {best_loss:.4f}")
        threshold = self.adam_smooth_minmax(best_loss)
        print(f"Threshold: {threshold:.4f} was chosen for best val loss of {best_loss:.4f}")
        cps_to_avg = [path for loss, path in zip(self.val_losses, self.cp_paths) if loss <= threshold]
        if len(cps_to_avg) < 1:
            print(f"⚠️ No checkpoints selected for averaging, skipping")
            return
        self.to_load_dir = self.avg_dir
        print(f"📊 Averaging {len(cps_to_avg)} checkpoints:")
        for cp_path in cps_to_avg:
            cp_idx = self.cp_paths.index(cp_path)
            marker = " 🏆" if cp_idx == best_epoch else ""
            print(f"- {basename(cp_path)} (val_loss={self.val_losses[cp_idx]:.4f}){marker}")

        averaged_weights = self.average_lora_checkpoints(cps_to_avg)
        averaged_checkpoint_path = join(self.save_dir, "checkpoint_averaged.pt")
        self.save_averaged_lora_checkpoint(base_checkpoint_path=cps_to_avg[0], averaged_lora_weights=averaged_weights,
                                           output_path=averaged_checkpoint_path)

        # Load averaged weights into model and save as a model
        model.load_state_dict(torch.load(averaged_checkpoint_path)['model_dict'], strict=False)
        averaged_model_dir = join(self.save_dir, "averaged_model")
        model.save_pretrained(averaged_model_dir)
        print(f"✅ Saved averaged model to {averaged_model_dir}")
        avg_val_loss, avg_val_metric = evaluator(val_loader)
        print(f"📈 Averaged checkpoint || Val Loss: {avg_val_loss:.4f} || Val Metric: {avg_val_metric:.4f}")
        self.avg_metric = avg_val_metric

    @classmethod
    def adam_smooth_minmax(cls, x):
        # Thanks to Adam Nathaniel for the "MinMax" idea
        return x + 0.02 + 0.03 * (x / (x + 0.3))


    @classmethod
    def extract_lora_weights(cls, checkpoint_path: str) -> Dict[str, torch.Tensor]:
        """
        Extract only LoRA adapter weights from a checkpoint.

        LoRA weights are named like:
        - base_model.model.{module_name}.lora_A.weight
        - base_model.model.{module_name}.lora_B.weight
        """
        checkpoint = torch.load(checkpoint_path, map_location='cpu')
        model_dict = checkpoint['model_dict']

        lora_weights = {}
        for key, value in model_dict.items():
            if 'lora_A' in key or 'lora_B' in key:
                lora_weights[key] = value
        return lora_weights

    @classmethod
    def average_lora_checkpoints(cls, checkpoint_paths: List[str]) -> Dict[str, torch.Tensor]:
        """
        LoRA adds two low-rank matrices A and B for each adapter layer:
        ΔW = B @ A  # Update to original weight matrix W
        W_new = W_frozen + α * (B @ A)
        B_avg = (B₁ + B₂ + ... + Bₙ) / n
        A_avg = (A₁ + A₂ + ... + Aₙ) / n

        Average LoRA adapter weights across multiple checkpoints.We naively average A and B matrices separately.
        """
        all_lora_weights = [cls.extract_lora_weights(path) for path in checkpoint_paths]
        # Verify all checkpoints have the same keys
        keys = set(all_lora_weights[0].keys())
        for i, weights in enumerate(all_lora_weights[1:], 1):
            if set(weights.keys()) != keys:
                raise ValueError(f"Checkpoint {i} has different LoRA parameters!")
        averaged_weights = {}
        for key in keys:
            stacked = torch.stack([weights[key] for weights in all_lora_weights])
            averaged_weights[key] = stacked.mean(dim=0)
        return averaged_weights

    @classmethod
    def save_averaged_lora_checkpoint(cls, base_checkpoint_path: str, averaged_lora_weights: Dict[str, torch.Tensor],
                                      output_path: str):
        """
        Save a new checkpoint with averaged LoRA weights.

        Args:
            base_checkpoint_path: Path to one of the original checkpoints (for metadata)
            averaged_lora_weights: Dictionary of averaged LoRA weights
            output_path: Path to save the averaged checkpoint
        """
        # Load the base checkpoint to get structure and non-LoRA components
        checkpoint = torch.load(base_checkpoint_path, map_location='cpu')

        # Replace LoRA weights with averaged ones
        model_dict = checkpoint['model_dict']
        for key in averaged_lora_weights:
            model_dict[key] = averaged_lora_weights[key]
        checkpoint['model_dict'] = model_dict

        # Save the new checkpoint
        torch.save(checkpoint, output_path)
        print(f"💾 Saved averaged checkpoint to {output_path}")




import os
import random
from typing import List

import numpy as np
import torch
from torch import Tensor

from tabstar.constants import SEED


def fix_seed(seed: int = SEED):
    if seed is None:
        return
    random.seed(seed)
    # Set a fixed value for the hash seed
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        # When running on the CuDNN backend, two further options must be set
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False


def concat_predictions(y_pred: List[Tensor]) -> np.ndarray:
    return np.concatenate([p.cpu().detach().numpy() for p in y_pred])


from pandas import Series
from sklearn.preprocessing import StandardScaler

from tabstar.preprocessing.nulls import get_invalid_indices

Z_MAX_ABS_VAL = 3


def fit_standard_scaler(s: Series) -> StandardScaler:
    s = s.copy().dropna()
    scaler = StandardScaler()
    scaler.fit(s.values.reshape(-1, 1))
    return scaler

def transform_clipped_z_scores(s: Series, scaler: StandardScaler, allow_null: bool = False) -> Series:
    invalid = get_invalid_indices(s)
    s = s.copy()
    if allow_null:
        s = s.fillna(s.mean())
    s_val = scaler.transform(s.values.reshape(-1, 1)).flatten()
    s_val = s_val.clip(-Z_MAX_ABS_VAL, Z_MAX_ABS_VAL)
    if invalid:
        s_val[sorted(invalid)] = 0
    s = Series(s_val, index=s.index, name=s.name)
    return s


import numpy as np
import torch
from torch import Tensor
from transformers import AutoTokenizer, AutoModel, PreTrainedModel
from tabstar.arch.config import TabStarConfig, E5_SMALL
from tabstar.arch.interaction import InteractionEncoder
from tabstar.arch.fusion import NumericalFusion
from tabstar.arch.prediction import PredictionHead
from tabstar.training.devices import clear_cuda_cache


class TabStarModel(PreTrainedModel):
    config_class = TabStarConfig

    def __init__(self, config: TabStarConfig):
        super().__init__(config)
        self.text_encoder = AutoModel.from_pretrained(E5_SMALL)
        self.tokenizer = AutoTokenizer.from_pretrained(E5_SMALL)
        self.numerical_fusion = NumericalFusion()
        self.tabular_encoder = InteractionEncoder()
        self.cls_head = PredictionHead()
        self.reg_head = PredictionHead()
        self.post_init()

    def forward(self, x_txt: np.ndarray, x_num: np.ndarray, d_output: int) -> Tensor:
        textual_embeddings = self.get_textual_embedding(x_txt)
        if not isinstance(x_num, Tensor):
            # TODO: this is a bug, it should always be a Tensor
            x_num = torch.tensor(x_num, dtype=textual_embeddings.dtype, device=textual_embeddings.device)
        embeddings = self.numerical_fusion(textual_embeddings=textual_embeddings, x_num=x_num)
        encoded = self.tabular_encoder(embeddings)
        target_tokens = encoded[:, :d_output]
        if d_output == 1:
            target_scores = self.reg_head(target_tokens)
        else:
            target_scores = self.cls_head(target_tokens)
        target_scores = target_scores.squeeze(dim=-1)
        assert tuple(target_scores.shape) == (x_txt.shape[0], d_output)
        return target_scores

    def get_textual_embedding(self, x_txt: np.array) -> Tensor:
        text_batch_size = 128
        while text_batch_size > 1:
            try:
                return self.get_textual_embedding_in_batches(x_txt, text_batch_size=text_batch_size)
            except torch.cuda.OutOfMemoryError:
                text_batch_size //= 2
                clear_cuda_cache()
                print(f"🤯 Reducing batch size to {text_batch_size} due to OOM")
        raise RuntimeError(f"🤯 OOM even with batch size 1!")

    def get_textual_embedding_in_batches(self, x_txt: np.array, text_batch_size: int) -> Tensor:
        # Get unique texts and mapping indices
        unique_texts, inverse_indices = np.unique(x_txt, return_inverse=True)
        num_unique_texts = len(unique_texts)
        embeddings = []
        for i in range(0, num_unique_texts, text_batch_size):
            batch_texts = unique_texts[i:i + text_batch_size].tolist()
            inputs = self.tokenizer(batch_texts, padding=True, return_tensors='pt', truncation=True)
            inputs = {k: v.to(self.device) for k, v in inputs.items()}
            outputs = self.text_encoder(**inputs)
            # Take the [CLS] token representation
            embeddings.append(outputs.last_hidden_state[:, 0, :])
        embeddings = torch.cat(embeddings, dim=0)
        inverse_indices = torch.tensor(inverse_indices, dtype=torch.long, device=embeddings.device)
        # Map the unique embeddings back to the original positions and reshape to match the original dimension
        batch_size, seq_len = x_txt.shape
        embeddings = embeddings[inverse_indices].view(batch_size, seq_len, -1)
        if not tuple(embeddings.shape) == (batch_size, seq_len, self.config.d_model):
            raise RuntimeError(f"Unexpected embedding shape: {embeddings.shape}")
        return embeddings



from tabstar.training.utils import fix_seed
from tabstar_paper.pretraining.seeds import py_state_to_json, py_state_from_json, np_state_to_json, np_state_from_json, \
    torch_state_to_json, torch_state_from_json


def test_py_seeds():
    fix_seed(0)
    py = py_state_to_json()
    py = py_state_from_json(py)
    assert py[0] == 3
    assert py[1][:5] == (2147483648, 766982754, 497961170, 3952298588, 2331775348)
    assert py[2] is None

def test_np_seeds():
    fix_seed(0)
    np_state = np_state_to_json()
    np_state = np_state_from_json(np_state)
    assert np_state[0] == 'MT19937'
    assert np_state[1].shape == (624,)
    assert np_state[1].tolist()[:5] == [0, 1, 1812433255, 1900727105, 1208447044]
    assert np_state[2] == 624
    assert np_state[3] == 0
    assert np_state[4] == 0.0


def test_torch_seeds():
    fix_seed(1306)
    torch_state = torch_state_to_json()
    torch_state = torch_state_from_json(torch_state)
    assert torch_state.tolist()[:3] == [26, 5, 0]



from dataclasses import dataclass
from typing import Optional, Dict, List, Union

import numpy as np
from pandas import DataFrame, Series
from sklearn.preprocessing import StandardScaler, QuantileTransformer, LabelEncoder
from skrub import DatetimeEncoder

from tabstar.preprocessing.binning import fit_numerical_bins, transform_numerical_bins
from tabstar.preprocessing.dates import fit_date_encoders, transform_date_features
from tabstar.preprocessing.feat_types import detect_numerical_features, transform_feature_types
from tabstar.preprocessing.nulls import raise_if_null_target
from tabstar.preprocessing.scaler import fit_standard_scaler, transform_clipped_z_scores
from tabstar.preprocessing.sparse import densify_objects
from tabstar.preprocessing.target import fit_preprocess_y, transform_preprocess_y
from tabstar.preprocessing.texts import replace_column_names
from tabstar.preprocessing.verbalize import prepend_target_tokens, verbalize_textual_features


@dataclass
class TabSTARData:
    d_output: int
    x_txt: Union[DataFrame, np.ndarray]
    x_num: np.ndarray
    y: Optional[Series] = None

    def __len__(self) -> int:
        return len(self.x_txt)

class TabSTARVerbalizer:
    def __init__(self, is_cls: bool, verbose: bool = False):
        self.is_cls = is_cls
        self.verbose = verbose
        self.date_transformers: Dict[str, DatetimeEncoder] = {}
        self.numerical_transformers: Dict[str, StandardScaler] = {}
        self.semantic_transformers: Dict[str, QuantileTransformer] = {}
        self.target_transformer: Optional[Union[LabelEncoder, StandardScaler]] = None
        self.d_output: Optional[int] = None
        self.y_name: Optional[str] = None
        self.y_values: Optional[List[str]] = None
        self.constant_columns: List[str] = []

    def fit(self, X, y):
        if len(X.columns) > 200:
            print("⚠️ Warning: More than 200 columns detected. This will probably lead to memory issues.")
        x = X.copy()
        y = y.copy()
        raise_if_null_target(y)
        self.assert_no_duplicate_columns(x)
        x, y = densify_objects(x=x, y=y)
        self.date_transformers = fit_date_encoders(x=x)
        self.vprint(f"📅 Detected {len(self.date_transformers)} date features: {sorted(self.date_transformers)}")
        x = transform_date_features(x=x, date_transformers=self.date_transformers)
        x, y = replace_column_names(x=x, y=y)
        numerical_features = detect_numerical_features(x)
        self.vprint(f"🔢 Detected {len(numerical_features)} numerical features: {sorted(numerical_features)}")
        text_features = [col for col in x.columns if col not in numerical_features]
        self.vprint(f"📝 Detected {len(text_features)} textual features: {sorted(text_features)}")
        x = transform_feature_types(x=x, numerical_features=numerical_features)
        self.target_transformer = fit_preprocess_y(y=y, is_cls=self.is_cls)
        if self.is_cls:
            self.d_output = len(self.target_transformer.classes_)
        else:
            self.d_output = 1
        self.constant_columns = [col for col in x.columns if x[col].nunique() == 1]
        for col in numerical_features:
            if col in self.constant_columns:
                continue
            self.numerical_transformers[col] = fit_standard_scaler(s=x[col])
            self.semantic_transformers[col] = fit_numerical_bins(s=x[col])
        self.y_name = str(y.name)
        if self.is_cls:
            self.y_values = sorted(self.target_transformer.classes_)

    def transform(self, x: DataFrame, y: Optional[Series]) -> TabSTARData:
        x = x.copy()
        if y is not None:
            y = y.copy()
        self.assert_no_duplicate_columns(x)
        x, y = densify_objects(x=x, y=y)
        x = transform_date_features(x=x, date_transformers=self.date_transformers)
        x, y = replace_column_names(x=x, y=y)
        num_cols = sorted(self.numerical_transformers)
        x = transform_feature_types(x=x, numerical_features=set(num_cols))
        y = self.transform_target(y=y)
        x = verbalize_textual_features(x=x)
        x = x.drop(columns=self.constant_columns, errors='ignore')
        x = prepend_target_tokens(x=x, y_name=self.y_name, y_values=self.y_values)
        text_cols = [col for col in x.columns if col not in num_cols]
        x_txt = x[text_cols + num_cols].copy()
        # x_num will hold the numerical features transformed to z-scores, and zero otherwise
        x_num = np.zeros(shape=x.shape, dtype=np.float32)
        for col in num_cols:
            x_txt[col] = transform_numerical_bins(s=x[col], scaler=self.semantic_transformers[col])
            idx = x_txt.columns.get_loc(col)
            s_num = transform_clipped_z_scores(s=x[col], scaler=self.numerical_transformers[col], allow_null=True)
            x_num[:, idx] = s_num.to_numpy()
        x_txt = x_txt.to_numpy()
        data = TabSTARData(d_output=self.d_output, x_txt=x_txt, x_num=x_num, y=y)
        return data

    def transform_target(self, y: Optional[Series]) -> Optional[Union[Series, np.ndarray]]:
        if y is None:
            return None
        y = y.copy()
        raise_if_null_target(y)
        return transform_preprocess_y(y=y, scaler=self.target_transformer)

    def inverse_transform_target(self, y):
        y = y.copy()
        assert isinstance(self.target_transformer, StandardScaler)
        return self.target_transformer.inverse_transform(y)

    @staticmethod
    def assert_no_duplicate_columns(x: DataFrame):
        if len(set(x.columns)) != len(x.columns):
            raise ValueError("Duplicate column names found in DataFrame!")

    def vprint(self, s: str):
        if self.verbose:
            print(s)



import torch
import torch.nn as nn

from tabstar.arch.config import D_MODEL


class InteractionEncoder(nn.Module):
    def __init__(self, num_layers: int = 6, d_model: int = D_MODEL, num_heads_factor: int = 64,
                 ffn_d_hidden_multiplier: int = 4, dropout: float = 0.1):
        super().__init__()
        dim_feedforward = d_model * ffn_d_hidden_multiplier
        num_heads = d_model // num_heads_factor
        encoder_layer = nn.TransformerEncoderLayer(
                d_model=d_model,
                nhead=num_heads,
                dim_feedforward=dim_feedforward,
                dropout=dropout,
                activation='relu',
                batch_first=True,
                norm_first=True
            )
        self.encoder = nn.TransformerEncoder(encoder_layer, num_layers=num_layers, enable_nested_tensor=False)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.encoder(x)



from typing import List

import numpy as np
from pandas import Series
from sklearn.preprocessing import QuantileTransformer

from tabstar.preprocessing.nulls import get_invalid_indices, MISSING_VALUE
from tabstar.preprocessing.verbalize import verbalize_feature

VERBALIZED_QUANTILE_BINS = 10

def fit_numerical_bins(s: Series) -> QuantileTransformer:
    s = s.copy().dropna()
    scaler = QuantileTransformer(output_distribution='uniform',
                                 n_quantiles=min(1000, len(s)),
                                 subsample=1000000000,
                                 random_state=0)
    scaler.fit(s.values.reshape(-1, 1))
    return scaler


def transform_numerical_bins(s: Series, scaler: QuantileTransformer) -> Series:
    invalid_indices = get_invalid_indices(s)
    quantile_levels = np.linspace(0, 1, VERBALIZED_QUANTILE_BINS + 1)
    boundaries = scaler.inverse_transform(quantile_levels.reshape(-1, 1)).flatten()
    assert len(boundaries) == VERBALIZED_QUANTILE_BINS + 1
    verbalized_bins = verbalize_bins(boundaries)
    bin_index = np.digitize(s, boundaries)
    verbalized = [verbalized_bins[i] for i in bin_index]
    for idx in invalid_indices:
        verbalized[idx] = MISSING_VALUE
    verbalized = [verbalize_feature(col=str(s.name), value=v) for v in verbalized]
    s = Series(verbalized, index=s.index, name=s.name)
    return s


def verbalize_bins(boundaries: np.array) -> List[str]:
    # TODO: this can become a bit ugly with high-precision numbers, or relatively-discrete numerical values
    boundaries = [format_float(b) for b in boundaries]
    first = f"Lower than {min(boundaries)} (Quantile 0%)"
    last = f"Higher than {max(boundaries)} (Quantile 100%)"
    bins = []
    for i, b in enumerate(boundaries[:-1]):
        r = f"{b} to {boundaries[i + 1]}"
        low = i * VERBALIZED_QUANTILE_BINS
        high = (i + 1) * VERBALIZED_QUANTILE_BINS
        q = f"(Quantile {low} - {high}%)"
        bins.append(f"{r} {q}")
    assert len(bins) == VERBALIZED_QUANTILE_BINS == len(boundaries) - 1
    bins = [first] + bins + [last]
    return bins


def format_float(num: float) -> str:
    rounded = round(num, 4)
    if rounded.is_integer():
        return str(int(rounded))
    formatted = f"{rounded:.4f}"
    formatted = formatted.rstrip("0").rstrip(".")
    return formatted


from typing import Any, Dict

import pandas as pd
from pandas import Series, DataFrame
from pandas.core.dtypes.common import is_datetime64_any_dtype
from skrub import DatetimeEncoder


def transform_date_features(x: DataFrame, date_transformers: Dict[str, DatetimeEncoder]) -> DataFrame:
    rows = x.shape[0]
    for col, dt_encoder in date_transformers.items():
        s = series_to_dt(s=x[col])
        dt_df = dt_encoder.transform(s)
        dt_df.index = x.index
        x = x.drop(columns=[col])
        x = pd.concat([x, dt_df], axis=1)
        if x.shape[0] != rows:
            raise ValueError(f"Row mismatch after transforming date column {col}: expected {rows}, got {x.shape}")
    return x

def fit_date_encoders(x: DataFrame) -> Dict[str, DatetimeEncoder]:
    date_encoders = {}
    date_columns = [str(col) for col, dtype in x.dtypes.items() if is_datetime64_any_dtype(dtype)]
    for col in date_columns:
        dt_s = series_to_dt(s=x[col])
        # Adds: "year", "month", "day", "hour", "total_seconds", "weekday"
        encoder = DatetimeEncoder(add_weekday=True, add_total_seconds=True)
        encoder.fit(dt_s)
        date_encoders[col] = encoder
    return date_encoders

def series_to_dt(s: Series) -> Series:
    # TODO: do we want to handle missing values here?
    s = s.apply(_clean_dirty_date)
    dt_s = pd.to_datetime(s, errors='coerce')
    dt_s = dt_s.apply(_remove_timezone)
    return dt_s


def _remove_timezone(dt):
    if pd.notnull(dt) and getattr(dt, 'tzinfo', None) is not None:
        return dt.tz_localize(None)
    return dt


def _clean_dirty_date(s: Any) -> Any:
    if isinstance(s, str):
        s = s.replace('"', "")
    return s


from typing import Optional

TEXT2FOLD = {
"REG_SOCIAL_ANIME_PLANET_RATING": 0,
"REG_CONSUMER_JC_PENNEY_PRODUCT_PRICE": 0,
"MUL_CONSUMER_PRODUCT_SENTIMENT": 0,
"MUL_SOCIAL_NEWS_CHANNEL_CATEGORY": 0,
"REG_HOUSES_CALIFORNIA_PRICES_2020": 0,
"REG_CONSUMER_BABIES_R_US_PRICES": 0,
"BIN_SOCIAL_JIGSAW_TOXICITY": 0,
"REG_SOCIAL_MUSEUMS_US_REVENUES": 0,
"REG_PROFESSIONAL_EMPLOYEE_RENUMERATION_VANCOUBER": 0,
"REG_SOCIAL_SPOTIFY_POPULARITY": 0,
"MUL_SOCIAL_SPOTIFY_GENRES": 0,             # For TabSTAR-v2's Text evaluation
"REG_FOOD_RAMEN_RATINGS_2022": 1,
"REG_SOCIAL_FILMTV_MOVIE_RATING_ITALY": 1,
"REG_CONSUMER_AMERICAN_EAGLE_PRICES": 1,
"REG_PROFESSIONAL_EMPLOYEE_SALARY_MONTGOMERY": 1,
"MUL_CONSUMER_WOMEN_ECOMMERCE_CLOTHING_REVIEW": 1,
"REG_SPORTS_FIFA22_WAGES": 1,
"REG_FOOD_WINE_VIVINO_SPAIN": 1,
"MUL_FOOD_WINE_REVIEW": 1,
"REG_PROFESSIONAL_COMPANY_EMPLOYEES_SIZE": 1,
"MUL_FOOD_YELP_REVIEWS": 1,
"REG_CONSUMER_BIKE_PRICE_BIKEWALE": 2,
"REG_TRANSPORTATION_USED_CAR_PAKISTAN": 2,
"MUL_HOUSES_MELBOURNE_AIRBNB": 2,
"BIN_PROFESSIONAL_KICKSTARTER_FUNDING": 2,
"MUL_FOOD_MICHELIN_GUIDE_RESTAURANTS": 2,
"REG_FOOD_ALCOHOL_WIKILIQ_PRICES": 2,
"REG_FOOD_COFFEE_REVIEW": 2,
"BIN_SOCIAL_IMDB_GENRE_PREDICTION": 2,
"REG_SOCIAL_KOREAN_DRAMA": 2,
"REG_FOOD_BEER_RATINGS": 2,
"REG_SOCIAL_BOOK_READABILITY_CLEAR": 3,
"MUL_SOCIAL_GOOGLE_QA_TYPE_REASON": 3,
"REG_SOCIAL_VIDEO_GAMES_SALES": 3,
"REG_SOCIAL_MOVIES_DATASET_REVENUE": 3,
"REG_FOOD_WINE_POLISH_MARKET_PRICES": 3,
"REG_FOOD_CHOCOLATE_BAR_RATINGS": 3,
"REG_TRANSPORTATION_USED_CAR_MERCEDES_BENZ_ITALY": 3,
"REG_CONSUMER_BOOK_PRICE_PREDICTION": 3,
"REG_TRANSPORTATION_USED_CAR_SAUDI_ARABIA": 3,
"MUL_TRANSPORTATION_US_ACCIDENTS_MARCH23": 3,
"REG_FOOD_ZOMATO_RESTAURANTS": 4,
"REG_SOCIAL_MOVIES_ROTTEN_TOMATOES": 4,
"REG_CONSUMER_CAR_PRICE_CARDEKHO": 4,
"REG_SPORTS_NBA_DRAFT_VALUE_OVER_REPLACEMENT": 4,
"MUL_PROFESSIONAL_DATA_SCIENTIST_SALARY": 4,
"REG_SOCIAL_BOOKS_GOODREADS": 4,
"REG_PROFESSIONAL_ML_DS_AI_JOBS_SALARIES": 4,
"REG_CONSUMER_MERCARI_ONLINE_MARKETPLACE": 4,
"BIN_PROFESSIONAL_FAKE_JOB_POSTING": 4,
"REG_PROFESSIONAL_SCIMAGOJR_ACADEMIC_IMPACT": 4,
}


PRETRAIN2FOLD = {
"MUL_ANONYM_FABERT": 0,
"BIN_CONSUMER_DRESS_DALES": 0,
"REG_SOCIAL_OCCUPATION_MOBILITY_SOCMOB": 0,
"BIN_HEALTHCARE_LIVER_INDIAN_ILPD": 0,
"BIN_NATURE_OZONE_LEVEL": 0,
"MUL_SOCIAL_BIAS_FRAMES": 0,
"BIN_COMPUTERS_PC4_CODE_DEFECTIONS": 0,
"REG_SPORTS_FOOTBALL_MANAGER_IMPORTANT_MATCHES": 0,
"MUL_SOCIAL_OKCUPID_DATING_JOB_STEM": 0,
"MUL_COMPUTERS_IMAGE_OPT_DIGITS": 0,
"BIN_ANONYM_MADELONE": 0,
"MUL_HEALTHCARE_HYPOTHYROID": 0,
"MUL_SPORTS_ACTIVITY_LOCALIZED_LDPA": 0,
"MUL_SCIENCE_ANNEAL_CHEMICAL": 0,
"REG_COMPUTERS_META_LEVEL_LEARNING": 0,
"REG_HOUSES_PERTH": 0,
"MUL_CONSUMER_ELECTRICITY_TAMILNADU": 0,
"REG_HOUSES_SEATTLE_KINGS_COUNTY": 0,
"REG_HOUSES_BOSTON_PRICE_NOMINAL": 0,
"MUL_SCIENCE_VOWEL_SPEAKER_RECOGNITION": 0,
"BIN_CONSUMER_MOBILE_CHURN": 0,
"REG_TRANSPORTATION_US_BIKE_SHARING_DEMAND": 0,
"BIN_ANONYM_JASMINE": 0,
"BIN_GEOGRAPHY_NOMAO_SEARCH_ENGINE": 0,
"REG_COMPUTERS_ROBOT_SARCOS": 0,
"REG_SCIENCE_SIMULATION_2D_PLANES": 0,
"BIN_SPORTS_RUN_OR_WALK": 0,
"REG_ANONYM_BUZZ_IN_SOCIAL_MEDIA_TWITTER": 0,
"BIN_SOCIAL_COMPASS_TWO_YEARS_OFFEND": 0,
"MUL_FINANCIAL_P2P_LOAN": 0,
"MUL_FINANCIAL_CREDIT_RISK_CHINA": 0,
"REG_NATURE_WIND_SPEED_IRELAND": 0,
"MUL_ANONYM_SYNTHETIC_CONTROL": 0,
"REG_HOUSES_GERMAN_HOUSES": 0,
"BIN_FINANCIAL_BANK_ACCOUNT_FRAUD_BAF": 0,
"MUL_GENETICS_ACP_BREAST_CANCER": 0,
"REG_ANONYM_PLANTS_TEXTURE": 0,
"REG_NATURE_FOREST_FIRES": 0,
"MUL_ANONYM_PENDIGITS": 0,
"BIN_FINANCIAL_ACCOUNT_DEBT_SOUTH_AFRICA": 0,
"BIN_COMPUTERS_KC1_CODE_DEFECTIONS": 0,
"BIN_COMPUTERS_JM1_CODE_DEFECTIONS": 0,         # For TabSTAR-v2's TabArena evaluation
"BIN_TRANSPORTATION_ROAD_SAFETY_GENDER": 0,
"BIN_ANONYM_GINA": 0,
"REG_ANONYM_YPROP": 0,
"MUL_TRANSPORTATION_CAR_STUDENT_INSURANCE_RISK": 0,
"BIN_FINANCIAL_BANK_MARKETING": 0,
"MUL_TRANSPORTATION_VEHICLE_SILHOUETTE": 0,
"REG_HOUSES_BOSTON_HOUSE": 0,
"BIN_CONSUMER_CLICK_PREDICTION": 0,
"MUL_SCIENCE_DRUG_DIRECTORY_FDA_UNAPPROVED": 0,
"MUL_NATURE_FOREST_COVERTYPE": 0,
"REG_COMPUTERS_CPU_ACTIVITY": 0,
"BIN_HEALTHCARE_GLIOMA_BRAIN_TUMOR": 0,
"BIN_ANONYM_AUTOUNIV_AU1": 0,
"BIN_COMPUTERS_PHISHING_URL_WEBSITES": 0,
"BIN_HEALTHCARE_THYROID_CANCER": 0,
"BIN_PROFESSIONAL_LICD_LABOR_RIGHTS": 0,
"MUL_TRANSPORTATION_CAR_ORIGIN_COUNTRY": 0,
"BIN_SOCIAL_SPEED_DATING": 0,
"MUL_ANONYM_ZERNIKE_MFEAT": 0,
"BIN_HEALTHCARE_HEART_DISEASE": 0,
"BIN_ANONYM_MAMMOGRAPHY": 0,
"MUL_HEALTHCARE_CANCER_CERVICAL": 0,
"REG_SCIENCE_SULFUR": 0,
"BIN_CONSUMER_TOUR_TRAVEL_CHURN": 0,
"MUL_SOCIAL_GOT_SCRIPTS": 0,
"REG_COMPUTERS_ROBOT_KIN8NM": 0,
"BIN_SOCIAL_TIC_TAC_TOE": 0,
"MUL_INDUSTRIAL_STEEL_PLATES_FAULT": 0,
"BIN_SCIENCE_PARTICLE_HIGGS": 0,
"BIN_ANONYM_KDDCUP_09_UPSELLING": 0,
"BIN_HEALTHCARE_SEPSIS_PHYSIONET": 0,
"BIN_SOCIAL_TWITTER_DISASTER": 0,
"MUL_NATURE_SOYBEAN_DISEASE": 0,
"BIN_ANONYM_PHILIPPINE": 0,
"REG_PROFESSIONAL_STUDENT_PERFORMANCE_PORTUGAL": 0,
"MUL_PROFESSIONAL_TEACHER_ASSISTANT_EVALUATION_TAE": 1,
"MUL_SPORTS_BASEBALL_HALL_OF_FAME": 1,
"MUL_COMPUTERS_IMAGE_CIFAR10": 1,
"MUL_ANONYM_SPOKEN_ARABIC_DIGIT": 1,
"REG_COMPUTERS_YOUTUBE_VIDEO_TRANSCODING": 1,
"MUL_SOCIAL_WIKIPEDIA_TALK_LABELS_ATTACKS": 1,
"BIN_ANONYM_XD6": 1,
"REG_NATURE_MYANMAR_AIR_QUALITY": 1,
"MUL_TRANSPORTATION_TRAFFIC_ACCIDENTS_FARS": 1,
"MUL_ANONYM_HELENA": 1,
"MUL_HEALTHCARE_EGYPTIAN_SKULLS": 1,
"BIN_NATURE_COLIC_HORSES_SURGICAL_LESION": 1,
"MUL_PROFESSIONAL_NURSERY_APPLICATIONS_SLOVENIA": 1,
"REG_SCIENCE_GRID_STABILITY": 1,
"BIN_HEALTHCARE_MENTAL_HEALTH_TECH": 1,
"BIN_COMPUTERS_MOZILLA": 1,
"BIN_HEALTHCARE_CELLS_WDBC_WISCONSIN_BREAST_CANCER": 1,
"REG_SPORTS_NBA_2K20_PLAYERS_RATING": 1,
"BIN_FINANCIAL_CC_TAIWAN_CREDIT_DEFAULT": 1,
"BIN_NATURE_DISEASED_TREES_WILT": 1,
"BIN_HEALTHCARE_DIABETES_EARLY_STAGE": 1,
"MUL_COMPUTERS_IMAGE_SEGMENTATION": 1,
"MUL_HEALTHCARE_HEPATITIS_C_EGYPT": 1,
"MUL_COMPUTERS_IMAGE_GESTURE_PHASE_SEGMENTATION": 1,
"MUL_NATURE_AIR_QUALITY_POLLUTION": 1,
"REG_ANONYM_TOPO": 1,
"BIN_GENETICS_MUSK_MOLECULES": 1,
"BIN_ANONYM_ARCENE": 1,
"REG_HOUSES_IOWA_HOUSES_PRICES": 1,
"BIN_CONSUMER_ELECTRICITY_PRICE_TREND": 1,
"BIN_SPORTS_PRO_FOOTBALL_BETS_PROFB": 1,
"BIN_CONSUMER_ONLINE_SHOPPERS_PURCHASE_INTENTION": 1,
"MUL_ANONYM_FIRST_ORDER_THEOREM_PROVING": 1,
"BIN_COMPUTERS_PHISHING_WEBSITE_HUDDERSFIELD": 1,
"MUL_COMPUTERS_PHISHING_WEBSITE_HUDDERSFIELD": 1,         # For TabSTAR-v2's TabArena evaluation
"MUL_SCIENCE_MICE_PROTEIN": 1,
"BIN_TRANSPORTATION_ESTONIA_DISASTER_PASSENGERS": 1,
"MUL_NATURE_EUCALYPTUS_SEED": 1,
"BIN_TRANSPORTATION_TITANIC_SURVIVAL": 1,
"MUL_SOCIAL_SOCC_OPINION_COMMENTS_CORPUS_TOXICITY": 1,
"MUL_FINANCIAL_LOAN_IBRD": 1,
"MUL_CONSUMER_INTERNET_USAGE_PROFESSION": 1,
"REG_NATURE_CLIMATE_CHANGE_AGRICULTURE_FINANCIAL": 1,
"BIN_CONSUMER_CHURN_TELCO_EUROPA": 1,
"MUL_FOOD_WINE_QUALITY_CAT": 1,
"REG_FOOD_WINE_QUALITY": 1,                             # For TabSTAR-v2's TabArena evaluation
"BIN_ANONYM_TWONORM": 1,
"BIN_ANONYM_GAMETES": 1,
"REG_HOUSES_MIAMI": 1,
"BIN_SCIENCE_HIV_QSAR": 1,
"MUL_SCIENCE_HIV_QSAR": 1,                              # For TabSTAR-v2's TabArena evaluation
"REG_SOCIAL_STRIKES_PER_COUNTRY": 1,
"MUL_TRANSPORTATION_SHUTTLE_SPACE": 1,
"MUL_ANONYM_DNA": 1,
"REG_TRANSPORTATION_US_AIRPORT_PASSENGERS": 1,
"REG_COMPUTERS_PUMA_ROBOT_ARM": 1,
"MUL_HEALTHCARE_MENTAL_HEALTH_OCCUPATION": 1,
"REG_TRANSPORTATION_FLIGHT_SIMULATION_ELEVATORS": 1,
"BIN_SCIENCE_EYE_STATE_EEG": 1,
"BIN_ANONYM_NUMERAI_28_6": 1,
"REG_SOCIAL_OKCUPID_DATING_PROFILE_AGE": 1,
"REG_NATURE_NO2_POLLUTION_NORWAY": 1,
"REG_SCIENCE_AUCTION_VERIFICATION": 1,
"REG_NATURE_ABALONE_FISH_RINGS": 1,
"MUL_HEALTHCARE_BABY_SLEEP_STATE": 1,
"MUL_SCIENCE_VOWEL_JAPANESE": 1,
"REG_COMPUTERS_ALGO_RUNTIME_SAT11": 1,
"MUL_NATURE_SOLAR_FLARES": 1,
"REG_HOUSES_CALIFORNIA_HOUSES": 1,
"BIN_CONSUMER_WHOLESALE_CUSTOMERS_PORTUGAL": 1,
"REG_HEALTHCARE_DIABETES_SKLEARN": 1,
"MUL_SPORTS_CHESS_JUNGLE": 1,
"MUL_SCIENCE_WAVEFORM_500": 1,
"REG_SCIENCE_QSAR_TID_10980": 1,
"BIN_FINANCIAL_AUDIT_RISK": 1,
"MUL_SOCIAL_HOLISTIC_BIAS": 2,
"BIN_PROFESSIONAL_ACADEMIC_COLLEGE_SCORECARD": 2,
"REG_SCIENCE_AIRFOIL_SELF_NOISE": 2,
"REG_TRANSPORTATION_NAVAL_PROPULSION_PLANT": 2,
"REG_ANONYM_BANK_32NH": 2,
"REG_SCIENCE_WAVE_ENERGY": 2,
"REG_SCIENCE_SUPERCONDUCTIVITY": 2,
"MUL_COMPUTERS_IMAGE_ARTIFICIAL_CHARACTERS": 2,
"MUL_SOCIAL_US_ELECTIONS_SPEECHES": 2,
"MUL_ANONYM_LED24": 2,
"REG_SPORTS_BASEBALL_HITTER_SALARY": 2,
"BIN_NATURE_MUSHROOM_POISONOUS": 2,
"MUL_HEALTHCARE_HEART_CARDIOTOCOGRAPHY": 2,
"REG_HEALTHCARE_BODY_FAT": 2,
"REG_NATURE_EL_NINO_KDD": 2,
"REG_CONSUMER_MEDICAL_CHARGES": 2,
"REG_HOUSES_LISBON": 2,
"MUL_SCIENCE_BALANCE_SCALE": 2,
"MUL_FINANCIAL_CORPORATE_CREDIT_RATING_SYMBOL": 2,
"BIN_ANONYM_KDDCUP_09_APPETENCY": 2,
"REG_SPORTS_FIFA20_PLAYERS_VALUE": 2,
"REG_CONSUMER_DIAMONDS_PRICES": 2,
"BIN_FINANCIAL_BANK_CUSTOMER_CHURN_SHRUTIME": 2,
"BIN_ANONYM_ALBERT": 2,
"REG_FINANCIAL_STOCK_AEROSPACE": 2,
"MUL_HEALTHCARE_PBCSEQ_BILIARY_CIRRHOSIS": 2,
"BIN_FINANCIAL_LT_VEHICLE_LOAN": 2,
"REG_SCIENCE_PHYSIOCHEMICAL_PROTEIN": 2,
"MUL_ANONYM_JANNIS": 2,
"BIN_TRANSPORTATION_AIRLINES_DEPARTURE_DELAY": 2,
"REG_SPORTS_MONEYBALL": 2,
"BIN_HEALTHCARE_BREAST_CANCER_YUGOSLAVIA": 2,
"BIN_SOCIAL_EDUCATIONAL_TRANSITIONS_IRISH": 2,
"MUL_COMPUTERS_INTRUSION_DETECTION_KDD": 2,
"MUL_SOCIAL_BROWN_FROWN_CORPORA_COLLINS": 2,
"MUL_HEALTHCARE_APPENDICITIS_REGENSBURG_PEDIATRIC": 2,
"BIN_FINANCIAL_BANKRUPTCY_QUALITATIVE": 2,
"MUL_ANONYM_DILBERT": 2,
"BIN_HEALTHCARE_HEART_FAILURE": 2,
"BIN_CONSUMER_INTERNET_ADVERTISEMENTS": 2,
"REG_ANONYM_ALLSTATE_CLAIM_SEVERITY": 2,
"REG_SOCIAL_US_CRIME": 2,
"MUL_SOCIAL_HARRY_POTTER_FAN_FICTION": 2,
"REG_SOCIAL_COLLEGES_GRANTS": 2,
"REG_SCIENCE_ENERGY_EFFICIENCY": 2,
"BIN_ANONYM_AUSTRALIAN_CREDIT_APPROVAL": 2,
"BIN_ANONYM_OIL_SPILL": 2,
"MUL_SCIENCE_BATCH_CHORAL_HARMONY": 2,
"BIN_FINANCIAL_BANK_PERSONAL_LOAN_MODELING": 2,
"BIN_FINANCIAL_CC_FRAUD_TRANSACTION": 2,
"BIN_CONSUMER_BLASTCHAR_TELCOM_CHURN": 2,
"REG_FINANCIAL_INSURANCE_HEALTH_HOURS_WORKED_WIFE": 2,
"MUL_COMPUTERS_IMAGE_MNIST_JAPANESE_KUZUSHIJI_49": 2,
"BIN_ANONYM_SYLVINE": 2,
"BIN_FINANCIAL_CREDIT_GERMAN": 2,
"BIN_CONSUMER_HOTEL_REVIEW": 2,
"BIN_SOCIAL_HUMAN_CHOICE_PREDICTION_LM_GAMES": 2,
"BIN_SOCIAL_SPAM_EMAILS_SPAMBASE": 2,
"REG_FINANCIAL_CC_USER_SEGMENTATION": 2,
"MUL_COMPUTERS_PAGE_BLOCK_PARSING": 2,
"MUL_SCIENCE_QSAR_BIOCONCENTRATION": 2,
"MUL_NATURE_CJS_MAPLE_TREES": 2,
"MUL_ANONYM_VOLCANOES_VENUS": 2,
"BIN_FINANCIAL_ADULT_INCOME": 2,
"REG_TRANSPORTATION_NYC_TAXI_TRIP_DURATION": 2,
"BIN_PROFESSIONAL_AMAZON_EMPLOYEE_ACCESS": 2,
"MUL_SOCIAL_FUNPEDIA": 2,
"BIN_PROFESSIONAL_EMPLOYEE_IBM_ATTRITION": 2,
"BIN_ANONYM_PORTO_SEGURO": 2,
"BIN_ANONYM_CHRISTINE": 3,
"REG_SOCIAL_VOTING_SPACE_GEOGRAPHIC_ANALYSIS": 3,
"BIN_ANONYM_ADA": 3,
"BIN_ANONYM_KDDCUP_98_DIRECT_MAIL": 3,
"BIN_HEALTHCARE_BLOOD_TRANSFUSION": 3,
"MUL_SOCIAL_BOOKS_AUTHORSHIP_ANALCATDATA": 3,
"REG_TRANSPORTATION_NYC_TAXI_TIP": 3,
"MUL_TRANSPORTATION_CAR_ACCEPTABILITY": 3,
"REG_NATURE_FISH_TOXICITY": 3,
"BIN_SPORTS_NBA_SHOTS": 3,
"MUL_SCIENCE_MIMIC_ROUTE": 3,
"MUL_GENETICS_SPLICE_DNA": 3,
"MUL_COMPUTERS_IMAGE_MNIST_FASHION": 3,
"BIN_FINANCIAL_TVS_CREDIT_LOAN": 3,
"BIN_ANONYM_GUILLERMO": 3,
"REG_CONSUMER_BOOKING_HOTEL_REVIEW": 3,
"MUL_COMPUTERS_IMAGE_MNIST_DIGITS": 3,
"BIN_TRANSPORTATION_CAR_BAD_BUY_KICK": 3,
"BIN_FINANCIAL_FRAUD_DETECTION": 3,
"REG_ANONYM_DIONIS": 3,
"MUL_TRANSPORTATION_TRAFFIC_VIOLATION": 3,
"BIN_SCIENCE_MAGIC_TELESCOPE": 3,
"BIN_HEALTHCARE_BREAST_CANCER_WISCONSIN": 3,
"MUL_TRANSPORTATION_AIRLINES_TWITTER_SENTIMENT": 3,
"REG_SCIENCE_QSAR_TID_11": 3,
"BIN_HEALTHCARE_DIABETES_US130": 3,     # For TabSTAR-v2's TabArena evaluation
"MUL_HEALTHCARE_DIABETES_US130": 3,
"BIN_SOCIAL_POLICE_INCIDENTS_SAN_FRANCISCO": 3,
"MUL_ANONYM_MICRO_MASS": 3,
"MUL_ANONYM_ROBERT": 3,
"MUL_SOCIAL_SUPREME_COURT_ANALCATDATA": 3,
"BIN_HEALTHCARE_BONE_MARROW_TRANSPLANT_CHILDREN": 3,
"REG_HOUSES_HOUSEHOLD_MONTHLY_ELECTRICITY": 3,
"REG_TRANSPORTATION_CAR_GM_PRICE": 3,
"MUL_NATURE_YEAST_PROTEIN": 3,
"BIN_COMPUTERS_CPU_TOKYO1": 3,
"MUL_HEALTHCARE_AUDIOLOGY_DIAGNOSTIC": 3,
"MUL_COMPUTERS_IMAGE_LETTER_RECOGNITION": 3,
"MUL_SOCIAL_STACKOVERFLOW_POLARITY": 3,
"REG_ANONYM_SANTANDER_TRANSACTION_VALUE": 3,
"BIN_FINANCIAL_HOME_EQUITY_CREDIT": 3,
"BIN_CONSUMER_REASONER_RECOMMENDATION": 3,
"MUL_HEALTHCARE_WHITE_BLOOD_CELLS_WBC": 3,
"MUL_SOCIAL_DPBEDIA": 3,
"REG_PROFESSIONAL_CPS88_WAGES": 3,
"BIN_SOCIAL_POLITICS_US_CONGRESS_VOTES": 3,
"BIN_GENETICS_OVA_BREAST": 3,
"REG_CONSUMER_BLACK_FRIDAY": 3,
"REG_CONSUMER_AVOCADO_SALES": 3,
"BIN_SCIENCE_CLIMATE_MODEL_SIMULATION_CRASHES": 3,
"BIN_CONSUMER_CHURN_TELCO_NIGERIA": 3,
"BIN_SOCIAL_HATE_SPEECH_DATASET_DYNAMICALLY_GENERATED": 3,
"REG_ANONYM_GEOGRAPHICAL_ORIGIN_OF_MUSIC": 3,
"MUL_HEALTHCARE_CMC_CONTRACEPTIVE_WOMEN": 3,
"BIN_HEALTHCARE_DIABETES_CDC_DEMOGRAPHICS": 3,
"REG_NATURE_POLLEN_LUXEMBOURG": 3,
"REG_FOOD_WHISKY_SCOTCH_REVIEWS": 3,
"BIN_HEALTHCARE_DIABETES_RISK_FACTORS": 3,
"BIN_FINANCIAL_CREDIT_FICO_HELOC": 3,
"REG_FINANCIAL_INSURANCE_PREMIUM_DATA": 3,
"MUL_COMPUTERS_ROBOT_WALL_NAVIGATION": 3,
"BIN_PROFESSIONAL_EMPLOYEE_TURNOVER_TECHCO": 3,
"BIN_SPORTS_CHESS_KR_VS_KP": 3,
"MUL_HEALTHCARE_OBESITY_LEVELS": 3,
"MUL_ANONYM_CNAE": 3,
"BIN_HEALTHCARE_KIDNEY_CHRONIC_DISEASE": 3,
"BIN_INDUSTRIAL_CYLINDER_BANDING_PRINTING": 3,
"MUL_ANONYM_MFEAT_FACTORS": 3,
"REG_TRANSPORTATION_ZURICH_PUBLIC_TRANSPORT_DELAY": 4,
"MUL_FINANCIAL_CREDIT_SCORE_HZL": 4,
"MUL_HEALTHCARE_HEART_ARRHYTMIA": 4,
"MUL_ANONYM_ISOLET_LETTER_SPEECH_RECOGNITION": 4,
"BIN_HEALTHCARE_ALZHEIMER_HANDWRITE_DARWIN": 4,
"MUL_PROFESSIONAL_STUDENT_PERFORMANCE_ENTRANCE_EXAM": 4,
"BIN_FINANCIAL_HOME_CREDIT_DEFAULT_RISK": 4,
"BIN_ANONYM_SATELLITE": 4,
"REG_ANONYM_FAT_MEAT_TECATOR": 4,
"MUL_SPORTS_POKER_HAND": 4,
"BIN_FINANCIAL_CREDIT_GIVE_ME_SOME": 4,
"BIN_HEALTHCARE_INSURANCE_LEAD_PREDICTION": 4,
"BIN_ANONYM_RICARDO": 4,
"BIN_SCIENCE_WATER_TREATMENT": 4,
"REG_ANONYM_MERCEDES_BENZ_GREENER_MANUFACTURING": 4,
"BIN_CONSUMER_CHURN_TELEPHONY": 4,
"BIN_FINANCIAL_LOAN_STATUS_PREDICTION": 4,
"MUL_ANONYM_AMAZON_COMMERCE_REVIEWS": 4,
"MUL_COMPUTERS_IMAGE_GTSRB_GERMAN_TRAFFIC_SIGN": 4,
"REG_FOOD_WINE_JUDGEMENT_SENSORY": 4,
"REG_COMPUTERS_ALGO_RUNTIME_MIP2016": 4,
"BIN_ANONYM_BIORESPONSE": 4,
"BIN_SCIENCE_PARTICLE_MINIBOONE": 4,
"MUL_ANONYM_POL": 4,
"BIN_HEALTHCARE_BREAST_CANCER_MAMMOGRAM": 4,
"REG_HEALTHCARE_POLLUTION_LA_MORTALITY_RMFTSA": 4,
"BIN_CONSUMER_CHURN_TELCO_SOUTH_ASIA": 4,
"REG_ANONYM_HOUSE_16H": 4,
"REG_SPORTS_FIFA_PLAYERS_STATS": 4,
"REG_NATURE_POLLEN_GRAINS_SYNTH": 4,
"BIN_SCIENCE_QSAR_BIODEG": 4,
"BIN_ANONYM_MONKS_PROBLEM_2": 4,
"BIN_HEALTHCARE_CARDIOVASCULAR_DISEASE": 4,
"BIN_SOCIAL_GAMING_LEAGUE_OF_LEGENDS_DIAMOND": 4,
"REG_SCIENCE_SIMULATION_FRIED_FORMULA_PREDICT": 4,
"BIN_ANONYM_APS_FAILURE": 4,
"BIN_FINANCIAL_CREDIT_RISK_DEFAULT": 4,
"MUL_FINANCIAL_CORPORATE_CREDIT_RATING_CAT": 4,
"REG_TRANSPORTATION_SEOUL_BIKE_SHARING_DEMAND": 4,
"BIN_NATURE_PHONEME_SOUNDS": 4,
"MUL_HEALTHCARE_TEETH_DMFT_ANALCATDATA": 4,
"REG_ANONYM_YOLANDA": 4,
"BIN_SCIENCE_EYE_MOVEMENT": 4,
"REG_SCIENCE_SIMULATION_MV": 4,
"REG_SCIENCE_CONCRETE_COMPRESSIVE_STRENGTH": 4,
"MUL_ANONYM_FOURIER_IMAGE_COEFFICIENT": 4,
"REG_HOUSES_BRAZILIAN_HOUSES": 4,
"BIN_HEALTHCARE_LUNG_MALE_ARSENIC": 4,
"REG_TRANSPORTATION_NYC_BIKE_TRIP_DURATION": 4,
"BIN_ANONYM_MADELINE": 4,
"REG_CONSUMER_ONLINE_NEWS_POPULARITY": 4,
"MUL_COMPUTERS_IMAGE_INDIAN_PINES": 4,
"BIN_SCIENCE_ANTIGEN_AVIDA_HIL6": 4,
"MUL_COMPUTERS_META_STREAM_INTERVALS": 4,
"MUL_SPORTS_CONNECT4_GAME": 4,
"MUL_HEALTHCARE_LYMPH_YUGOSLAVIA": 4,
"BIN_SCIENCE_SIMULATION_HILL_VALLEY": 4,
"BIN_CONSUMER_NEWSPAPER_CHURN": 4,
"BIN_FINANCIAL_CC_FRAUD": 4,
"MUL_SPORTS_WALKING_ACTIVITY": 4,
"REG_COMPUTERS_GAMING_FRAMES_FPS_BENCHMARK": 4,
"REG_NATURE_QUAKE_RICHTER": 4,
"MUL_ANONYM_VOLKERT": 4,
"BIN_COMPUTERS_IMAGE_SKIN_SEGMENTATION": 4,
"BIN_COMPUTERS_IMAGE_BANK_NOTE_AUTHENTICATION": 4,
"REG_SPORTS_NBA_ALL_STAR": 4,
}


def get_tabstar_version(pretrain_dataset_or_path: Optional[str] = None) -> str:
    if pretrain_dataset_or_path is None:
        return "alana89/TabSTAR"
    if pretrain_dataset_or_path.startswith(("BIN_", "REG_", "MUL_")):
        tabstar_version = _get_tabstar_version_from_dataset(pretrain_dataset=pretrain_dataset_or_path)
        return f"alana89/{tabstar_version}"
    if isinstance(pretrain_dataset_or_path, str):
        return pretrain_dataset_or_path
    raise ValueError(f"Unknown pretrain_dataset_or_path: {pretrain_dataset_or_path}")


def _get_tabstar_version_from_dataset(pretrain_dataset: str) -> str:
    text_fold = TEXT2FOLD.get(pretrain_dataset)
    if text_fold is not None:
        return f"TabSTAR-paper-version-fold-k{text_fold}"
    pretrain_fold = PRETRAIN2FOLD.get(pretrain_dataset)
    if pretrain_fold is not None:
        return f"TabSTAR-eval-320-version-fold-k{pretrain_fold}"
    print(f"🤷‍♂️ Unknown dataset: {pretrain_dataset}, falling back to TabSTAR")
    return "TabSTAR"


from transformers import PretrainedConfig


D_MODEL = 384
E5_SMALL = 'intfloat/e5-small-v2'
WEIGHT_DECAY = 0.001

# TODO: this actually should include many arch parameters: /tabular/tabstar/params/config.py
class TabStarConfig(PretrainedConfig):
    model_type = "tabstar"

    def __init__(
        self,
        d_model: int = D_MODEL,
        embedding_model: str = E5_SMALL,
        weight_decay: float = WEIGHT_DECAY,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.d_model = d_model
        self.embedding_model = embedding_model
        self.weight_decay = weight_decay



from typing import Optional, Tuple, Union

import joblib
import numpy as np
import torch
from pandas import Series, DataFrame
from peft import PeftModel
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
from torch import softmax

from tabstar.arch.arch import TabStarModel
from tabstar.preprocessing.nulls import raise_if_null_target
from tabstar.preprocessing.splits import split_to_val
from tabstar.tabstar_datasets import get_tabstar_version
from tabstar.tabstar_verbalizer import TabSTARVerbalizer, TabSTARData
from tabstar.training.dataloader import get_dataloader
from tabstar.training.devices import get_device
from tabstar.training.hyperparams import LORA_LR, LORA_R, MAX_EPOCHS, FINETUNE_PATIENCE, LORA_BATCH, GLOBAL_BATCH
from tabstar.training.metrics import calculate_metric, Metrics
from tabstar.training.trainer import TabStarTrainer
from tabstar.training.utils import concat_predictions, fix_seed


class BaseTabSTAR:
    def __init__(self,
                 is_paper_version: bool = False,
                 lora_lr: float = LORA_LR,
                 lora_r: int = LORA_R,
                 lora_batch: int = LORA_BATCH,
                 global_batch: int = GLOBAL_BATCH,
                 max_epochs: int = MAX_EPOCHS,
                 patience: int = FINETUNE_PATIENCE,
                 verbose: bool = False,
                 device: Optional[Union[str,  torch.device]] = None,
                 random_state: Optional[int] = None,
                 time_limit: Optional[int] = None,
                 pretrain_dataset_or_path: Optional[str] = None,
                 keep_model: bool = True,
                 output_dir: Optional[str] = None,
                 metric_name: Optional[str] = None,
                 ):
        self.cp_average = not bool(is_paper_version)
        self.lora_lr = lora_lr
        self.lora_r = lora_r
        self.lora_batch = lora_batch
        self.global_batch = global_batch
        self.max_epochs = max_epochs
        self.patience = patience
        self.verbose = verbose
        self.preprocessor_: Optional[TabSTARVerbalizer] = None
        self.model_: Optional[PeftModel] = None
        self.random_state = random_state
        self.time_limit = time_limit
        self.keep_model = keep_model
        self.output_dir = output_dir
        self.metric_name = metric_name
        fix_seed(seed=self.random_state)
        self.device = get_device(device=device)
        print(f"🖥️ Using device: {self.device}")
        self.use_amp = bool(self.device.type == "cuda")
        self.model_version = get_tabstar_version(pretrain_dataset_or_path=pretrain_dataset_or_path)

    def fit(self, X: DataFrame, y: Series, x_val: Optional[DataFrame] = None, y_val: Optional[DataFrame] = None):
        if self.model_ is not None:
            raise ValueError("Model is already trained. Call fit() only once.")
        self.vprint(f"Fitting model on data with shapes: X={X.shape}, y={y.shape}")
        train_data, val_data = self._prepare_for_train(X, y, x_val, y_val)
        self.vprint(f"We have: {len(train_data)} training and {len(val_data)} validation samples.")
        trainer = TabStarTrainer(lora_lr=self.lora_lr,
                                 lora_r=self.lora_r,
                                 lora_batch=self.lora_batch,
                                 global_batch=self.global_batch,
                                 max_epochs=self.max_epochs,
                                 patience=self.patience,
                                 device=self.device,
                                 model_version=self.model_version,
                                 cp_average=self.cp_average,
                                 time_limit=self.time_limit,
                                 output_dir=self.output_dir,
                                 metric_name=self.metric_name)
        trainer.train(train_data, val_data)
        self.model_ = trainer.load_model()
        if not self.keep_model:
            trainer.delete_model()

    @staticmethod
    def download_base_model():
        TabStarModel.from_pretrained("alana89/TabSTAR")
        print(f"Downloaded base mode!")

    def predict(self, X):
        raise NotImplementedError("Must be implemented in subclass")

    @property
    def is_cls(self) -> bool:
        raise NotImplementedError("Must be implemented in subclass")

    def save(self, path: str):
        joblib.dump(self, path, compress=3)

    @classmethod
    def load(cls, path: str) -> 'BaseTabSTAR':
        return joblib.load(path)

    def _prepare_for_train(self, X, y, x_val: Optional[DataFrame], y_val: Optional[Series]) -> Tuple[TabSTARData, TabSTARData]:
        if not isinstance(X, DataFrame):
            raise ValueError("X must be a pandas DataFrame.")
        if not isinstance(y, Series):
            raise ValueError("y must be a pandas Series.")
        raise_if_null_target(y)
        self.vprint(f"Preparing data for training. X shape: {X.shape}, y shape: {y.shape}")
        if x_val is None and y_val is None:
            x_train, x_val, y_train, y_val = split_to_val(x=X, y=y, is_cls=self.is_cls)
            self.vprint(f"Split to validation set. Train has {len(x_train)} samples, validation has {len(x_val)} samples.")
        else:
            x_train = X.copy()
            y_train = y.copy()
            raise_if_null_target(y_val)
        if self.preprocessor_ is None:
            self.preprocessor_ = TabSTARVerbalizer(is_cls=self.is_cls, verbose=self.verbose)
            self.preprocessor_.fit(x_train, y_train)
        train_data = self.preprocessor_.transform(x_train, y_train)
        self.vprint(f"Transformed training data: {train_data.x_txt.shape=}, x_num shape: {train_data.x_num.shape=}")
        val_data = self.preprocessor_.transform(x_val, y_val)
        return train_data, val_data

    def _infer(self, X) -> np.ndarray:
        self.model_.eval()
        data = self.preprocessor_.transform(X, y=None)
        dataloader = get_dataloader(data, is_train=False, batch_size=128)
        predictions = []
        for data in dataloader:
            with torch.no_grad(), torch.autocast(device_type=self.device.type, enabled=self.use_amp):
                batch_predictions = self.model_(x_txt=data.x_txt, x_num=data.x_num, d_output=data.d_output)
            batch_predictions = batch_predictions.to(torch.float32)
            if self.is_cls:
                batch_predictions = softmax(batch_predictions, dim=1)
            predictions.append(batch_predictions)
        predictions = concat_predictions(predictions)
        return predictions

    def vprint(self, s: str):
        if self.verbose:
            print(s)

    def score(self, X, y) -> float:
        metrics = self.score_all_metrics(X=X, y=y)
        return metrics.score

    def score_all_metrics(self, X, y) -> Metrics:
        x = X.copy()
        y = y.copy()
        y_pred = self._infer(x)
        y_true = self.preprocessor_.transform_target(y)
        metrics = calculate_metric(y_true=y_true, y_pred=y_pred, d_output=self.preprocessor_.d_output)
        return metrics



class TabSTARClassifier(BaseTabSTAR, BaseEstimator, ClassifierMixin):

    def predict(self, X):
        if not isinstance(self.model_, PeftModel):
            raise ValueError("Model is not trained yet. Call fit() before predict().")
        predictions = self._infer(X)
        if predictions.ndim == 1:
            predictions = np.round(predictions)
        else:
            predictions = np.argmax(predictions, axis=1)
        label = self.preprocessor_.target_transformer.inverse_transform(predictions)
        return label

    def predict_proba(self, X):
        return self._infer(X)

    @property
    def is_cls(self) -> bool:
        return True

    @property
    def classes_(self) -> np.ndarray:
        if self.preprocessor_ is None or self.preprocessor_.y_values is None:
            raise ValueError("Model is not trained yet! Call fit() before accessing classes_.")
        return np.array(self.preprocessor_.y_values)


class TabSTARRegressor(BaseTabSTAR, BaseEstimator, RegressorMixin):

    def predict(self, X):
        if not isinstance(self.model_, PeftModel):
            raise ValueError("Model is not trained yet. Call fit() before predict().")
        z_scores = self._infer(X)
        y_pred = self.preprocessor_.inverse_transform_target(z_scores)
        return y_pred.flatten()

    @property
    def is_cls(self) -> bool:
        return False





from typing import List, Any, Optional, Set

import numpy as np
import pandas as pd
from pandas import Series

MISSING_VALUE = "Unknown Value"


def raise_if_null_target(y: Series):
    # TODO: do we want to allow throwing these away with a flag?
    # TODO: for multiclass, warn if any category is too rare?
    missing = y.isnull()
    y_missing = missing.sum()
    if y_missing > 0:
        raise ValueError(f"Target variable {y.name} has {y_missing} null values, please handle them before training.")


def get_invalid_indices(ls: Series) -> Set[int]:
    return {i for i, x in enumerate(ls) if _get_non_null_value(x) is None}


def get_valid_values(ls: Series) -> List:
    return [x for x in ls if _get_non_null_value(x) is not None]

def _get_non_null_value(x: Any) -> Optional[Any]:
    if isinstance(x, str):
        return x
    if pd.isna(x):
        return None
    if not np.isfinite(x):
        return None
    if pd.isnull(x):
        return None
    return x


def convert_numeric_with_missing(s: pd.Series, missing_value: Any) -> pd.Series:
    return s.apply(lambda x: x if x != missing_value else np.nan).astype(float)



from typing import Dict, Optional

from tabstar.training.metrics import METRICS_TO_MINIMIZE


class EarlyStopping:

    def __init__(self, patience: int, metric_name: Optional[str]):
        self.metric: float = float('-inf')
        self.loss: float = float('inf')
        self.failed: int = 0
        self.patience = patience
        self.maximize_metric = True
        if metric_name in METRICS_TO_MINIMIZE:
            self.maximize_metric = False
            self.metric = float('inf')

    def update_metric(self, metric: float) -> str:
        if self.maximize_metric:
            is_improve = bool(metric > self.metric)
        else:
            is_improve = bool(metric < self.metric)
        if is_improve:
            self.metric = metric
            self.failed = 0
            return " 🥇"
        else:
            self.failed += 1
            return f" 😓 [{self.failed} / {self.patience}]"

    def update_loss(self, loss: float) -> str:
        if loss < self.loss:
            self.loss = loss
            self.failed = 0
            return " 🥇"
        else:
            self.failed += 1
            return f" 😓 [{self.failed} / {self.patience}]"

    @property
    def is_best(self) -> bool:
        return self.failed == 0

    @property
    def should_stop(self) -> bool:
        return self.failed >= self.patience

    def state_dict(self) -> Dict:
        return {'metric': self.metric, 'failed': self.failed, 'patience': self.patience, 'loss': self.loss}

    def load_state_dict(self, state_dict: Dict):
        self.metric = state_dict['metric']
        self.loss = state_dict['loss']
        self.failed = state_dict['failed']
        self.patience = state_dict['patience']


import pandas as pd

from tabstar.preprocessing.splits import merge_splits

# TODO: AI assisted tests, need to understand better if they make sense
def _make_toy(overlap_indices=False):
    # Build data where y equals row-wise sum of x, so we can verify alignment after shuffling
    x1 = pd.DataFrame({"a": [1, 2, 3], "b": [10, 20, 30]})
    x2 = pd.DataFrame({"a": [4, 5], "b": [40, 50]})
    if overlap_indices:
        # Force duplicate labels across splits (0,1) to simulate a common pitfall
        x2.index = x1.index[:len(x2)]
    y1 = x1.sum(axis=1)  # 11, 22, 33
    y2 = x2.sum(axis=1)  # 44, 55
    return x1, x2, y1, y2

def test_merge_splits_unique_indices_aligns():
    x1, x2, y1, y2 = _make_toy(overlap_indices=False)
    x, y = merge_splits(x1, x2, y1, y2, random_state=42)
    # Recompute the per-row sum and ensure it equals shuffled y
    recomputed = x.sum(axis=1).reset_index(drop=True)
    pd.testing.assert_series_equal(recomputed, y.reset_index(drop=True), check_names=False)

def test_merge_splits_overlapping_indices_aligns():
    x1, x2, y1, y2 = _make_toy(overlap_indices=True)
    x, y = merge_splits(x1, x2, y1, y2, random_state=42)
    recomputed = x.sum(axis=1).reset_index(drop=True)
    pd.testing.assert_series_equal(recomputed, y.reset_index(drop=True), check_names=False)



import numpy as np
import pandas as pd
from pandas import Series

from tabstar.preprocessing.nulls import get_invalid_indices


def test_get_valid_indices():
    s = Series([1, 2, 3, np.nan, np.inf, 6])
    assert get_invalid_indices(s) == {3, 4}

    s = Series([None, 1, float("inf"), np.NaN])
    assert get_invalid_indices(s) == {0, 2, 3}

    s = [1, 2, None, np.nan, pd.NA, float("nan"), float("NaN"), float("inf"), np.inf, -np.inf]
    assert get_invalid_indices(s) == {2, 3, 4, 5, 6, 7, 8, 9}




from enum import Enum
from typing import Union


class OpenMLDatasetID(Enum):
    BIN_ANONYM_ADA = 41156
    BIN_ANONYM_ALBERT = 41147
    BIN_ANONYM_APS_FAILURE = 41138
    BIN_ANONYM_ARCENE = 41157
    BIN_ANONYM_AUSTRALIAN_CREDIT_APPROVAL = 40981
    BIN_ANONYM_AUTOUNIV_AU1 = 1547
    BIN_ANONYM_BIORESPONSE = 4134
    BIN_ANONYM_CHRISTINE = 41142
    BIN_ANONYM_GAMETES = 40650
    BIN_ANONYM_GINA = 41158
    BIN_ANONYM_GUILLERMO = 41159
    BIN_ANONYM_JASMINE = 41143
    BIN_ANONYM_KDDCUP_98_DIRECT_MAIL = 42343
    BIN_ANONYM_KDDCUP_09_APPETENCY = 1111
    BIN_ANONYM_KDDCUP_09_UPSELLING = 43072
    BIN_ANONYM_MADELINE = 41144
    BIN_ANONYM_MADELONE = 1485
    BIN_ANONYM_MAMMOGRAPHY = 310
    BIN_ANONYM_MONKS_PROBLEM_2 = 334
    BIN_ANONYM_NUMERAI_28_6 = 23517
    BIN_ANONYM_OIL_SPILL = 311
    BIN_ANONYM_PHILIPPINE = 41145
    BIN_ANONYM_PORTO_SEGURO = 42742
    BIN_ANONYM_RICARDO = 41161
    BIN_ANONYM_SATELLITE = 40900
    BIN_ANONYM_SYLVINE = 41146
    BIN_ANONYM_TWONORM = 1507
    BIN_ANONYM_XD6 = 40693
    BIN_COMPUTERS_CPU_TOKYO1 = 40705
    BIN_COMPUTERS_IMAGE_BANK_NOTE_AUTHENTICATION = 43466
    BIN_COMPUTERS_IMAGE_SKIN_SEGMENTATION = 1502
    BIN_COMPUTERS_MOZILLA = 1046
    BIN_COMPUTERS_KC1_CODE_DEFECTIONS = 1067
    BIN_COMPUTERS_PC4_CODE_DEFECTIONS = 1049
    BIN_COMPUTERS_PHISHING_URL_WEBSITES = 43622
    BIN_COMPUTERS_PHISHING_WEBSITE_HUDDERSFIELD = 4534
    BIN_CONSUMER_BLASTCHAR_TELCOM_CHURN = 46280
    BIN_CONSUMER_CHURN_TELCO_EUROPA = 44228
    BIN_CONSUMER_CHURN_TELCO_NIGERIA = 44230
    BIN_CONSUMER_CHURN_TELCO_SOUTH_ASIA = 44227
    BIN_CONSUMER_CHURN_TELEPHONY = 40701
    BIN_CONSUMER_CLICK_PREDICTION = 42733
    BIN_CONSUMER_DRESS_DALES = 23381
    BIN_CONSUMER_ELECTRICITY_PRICE_TREND = 44156
    BIN_CONSUMER_HOTEL_REVIEW = 43721
    BIN_CONSUMER_INTERNET_ADVERTISEMENTS = 40978
    BIN_CONSUMER_MOBILE_CHURN = 44231
    BIN_CONSUMER_NEWSPAPER_CHURN = 44226
    BIN_CONSUMER_ONLINE_SHOPPERS_PURCHASE_INTENTION = 45560
    BIN_CONSUMER_TOUR_TRAVEL_CHURN = 45545
    BIN_CONSUMER_WHOLESALE_CUSTOMERS_PORTUGAL = 1511
    BIN_FINANCIAL_ACCOUNT_DEBT_SOUTH_AFRICA = 43838
    BIN_FINANCIAL_ADULT_INCOME = 1590
    BIN_FINANCIAL_AUDIT_RISK = 42931
    BIN_FINANCIAL_BANK_ACCOUNT_FRAUD_BAF = 46646
    BIN_FINANCIAL_BANK_CUSTOMER_CHURN_SHRUTIME = 45062
    BIN_FINANCIAL_BANK_MARKETING = 1461
    BIN_FINANCIAL_BANK_PERSONAL_LOAN_MODELING = 43826
    BIN_FINANCIAL_BANKRUPTCY_QUALITATIVE = 1495
    BIN_FINANCIAL_CC_FRAUD = 45955
    BIN_FINANCIAL_CC_FRAUD_TRANSACTION = 46369
    BIN_FINANCIAL_CC_TAIWAN_CREDIT_DEFAULT = 43435
    BIN_FINANCIAL_CREDIT_FICO_HELOC = 45554
    BIN_FINANCIAL_CREDIT_GERMAN = 31
    BIN_FINANCIAL_CREDIT_GIVE_ME_SOME = 45577
    BIN_FINANCIAL_CREDIT_RISK_DEFAULT = 43454
    BIN_FINANCIAL_FRAUD_DETECTION = 46359
    BIN_FINANCIAL_HOME_CREDIT_DEFAULT_RISK = 45567
    BIN_FINANCIAL_HOME_EQUITY_CREDIT = 43337
    BIN_FINANCIAL_LOAN_STATUS_PREDICTION = 43595
    BIN_FINANCIAL_LT_VEHICLE_LOAN = 46430
    BIN_FINANCIAL_TVS_CREDIT_LOAN = 43743
    BIN_GENETICS_MUSK_MOLECULES = 1116
    BIN_GENETICS_OVA_BREAST = 1128
    BIN_GEOGRAPHY_NOMAO_SEARCH_ENGINE = 1486
    BIN_HEALTHCARE_ALZHEIMER_HANDWRITE_DARWIN = 46606
    BIN_HEALTHCARE_BLOOD_TRANSFUSION = 1464
    BIN_HEALTHCARE_BONE_MARROW_TRANSPLANT_CHILDREN = 46610
    BIN_HEALTHCARE_BREAST_CANCER_MAMMOGRAM = 43687
    BIN_HEALTHCARE_BREAST_CANCER_WISCONSIN = 15
    BIN_HEALTHCARE_BREAST_CANCER_YUGOSLAVIA = 13
    BIN_HEALTHCARE_CELLS_WDBC_WISCONSIN_BREAST_CANCER = 1510
    BIN_HEALTHCARE_CARDIOVASCULAR_DISEASE = 45547
    BIN_HEALTHCARE_DIABETES_CDC_DEMOGRAPHICS = 46598
    BIN_HEALTHCARE_DIABETES_EARLY_STAGE = 43643
    BIN_HEALTHCARE_DIABETES_RISK_FACTORS = 46254
    BIN_HEALTHCARE_GLIOMA_BRAIN_TUMOR = 46604
    BIN_HEALTHCARE_HEART_DISEASE = 43672
    BIN_HEALTHCARE_HEART_FAILURE = 45950
    BIN_HEALTHCARE_INSURANCE_LEAD_PREDICTION = 45064
    BIN_HEALTHCARE_KIDNEY_CHRONIC_DISEASE = 42972
    BIN_HEALTHCARE_LIVER_INDIAN_ILPD = 1480
    BIN_HEALTHCARE_LUNG_MALE_ARSENIC = 951
    BIN_HEALTHCARE_MENTAL_HEALTH_TECH = 46719
    BIN_HEALTHCARE_SEPSIS_PHYSIONET = 46677
    BIN_HEALTHCARE_THYROID_CANCER = 46605
    BIN_INDUSTRIAL_CYLINDER_BANDING_PRINTING = 6332
    BIN_NATURE_COLIC_HORSES_SURGICAL_LESION = 25
    BIN_NATURE_DISEASED_TREES_WILT = 40983
    BIN_NATURE_OZONE_LEVEL = 1487
    BIN_NATURE_MUSHROOM_POISONOUS = 24
    BIN_NATURE_PHONEME_SOUNDS = 1489
    BIN_PROFESSIONAL_ACADEMIC_COLLEGE_SCORECARD = 46674
    BIN_PROFESSIONAL_AMAZON_EMPLOYEE_ACCESS = 4135
    BIN_PROFESSIONAL_EMPLOYEE_IBM_ATTRITION = 43893
    BIN_PROFESSIONAL_EMPLOYEE_TURNOVER_TECHCO = 43551
    BIN_PROFESSIONAL_FAKE_JOB_POSTING = 46655
    BIN_PROFESSIONAL_KICKSTARTER_FUNDING = 46668
    BIN_PROFESSIONAL_LICD_LABOR_RIGHTS = 46640
    BIN_SCIENCE_ANTIGEN_AVIDA_HIL6 = 46648
    BIN_SCIENCE_CLIMATE_MODEL_SIMULATION_CRASHES = 40994
    BIN_SCIENCE_EYE_MOVEMENT = 44157
    BIN_SCIENCE_EYE_STATE_EEG = 1471
    BIN_SCIENCE_HIV_QSAR = 1039
    BIN_SCIENCE_MAGIC_TELESCOPE = 44125
    BIN_SCIENCE_PARTICLE_HIGGS = 42769
    BIN_SCIENCE_PARTICLE_MINIBOONE = 41150
    BIN_SCIENCE_QSAR_BIODEG = 1494
    BIN_SCIENCE_SIMULATION_HILL_VALLEY = 1479
    BIN_SCIENCE_WATER_TREATMENT = 940
    BIN_SOCIAL_COMPASS_TWO_YEARS_OFFEND = 45039
    BIN_SOCIAL_EDUCATIONAL_TRANSITIONS_IRISH = 451
    BIN_SOCIAL_IMDB_GENRE_PREDICTION = 46667
    BIN_SOCIAL_GAMING_LEAGUE_OF_LEGENDS_DIAMOND = 43635
    BIN_SOCIAL_HATE_SPEECH_DATASET_DYNAMICALLY_GENERATED = 46683
    BIN_SOCIAL_JIGSAW_TOXICITY = 46654
    BIN_SOCIAL_POLICE_INCIDENTS_SAN_FRANCISCO = 42732
    BIN_SOCIAL_POLITICS_US_CONGRESS_VOTES = 56
    BIN_SOCIAL_SPAM_EMAILS_SPAMBASE = 44
    BIN_SOCIAL_SPEED_DATING = 40536
    BIN_SOCIAL_TIC_TAC_TOE = 50
    BIN_SOCIAL_TWITTER_DISASTER = 43395
    BIN_SPORTS_CHESS_KR_VS_KP = 3
    BIN_SPORTS_NBA_SHOTS = 42806
    BIN_SPORTS_PRO_FOOTBALL_BETS_PROFB = 470
    BIN_SPORTS_RUN_OR_WALK = 40922
    BIN_TRANSPORTATION_AIRLINES_DEPARTURE_DELAY = 1169
    BIN_TRANSPORTATION_CAR_BAD_BUY_KICK = 41162
    BIN_TRANSPORTATION_ESTONIA_DISASTER_PASSENGERS = 43389
    BIN_TRANSPORTATION_ROAD_SAFETY_GENDER = 45038
    BIN_TRANSPORTATION_TITANIC_SURVIVAL = 40945
    MUL_ANONYM_AMAZON_COMMERCE_REVIEWS = 1457
    MUL_ANONYM_CNAE = 1468
    MUL_ANONYM_DILBERT = 41163
    MUL_ANONYM_DNA = 40670
    MUL_ANONYM_FABERT = 41164
    MUL_ANONYM_FIRST_ORDER_THEOREM_PROVING = 1475
    MUL_ANONYM_FOURIER_IMAGE_COEFFICIENT = 14
    MUL_ANONYM_HELENA = 41169
    MUL_ANONYM_ISOLET_LETTER_SPEECH_RECOGNITION = 300
    MUL_ANONYM_JANNIS = 41168
    MUL_ANONYM_LED24 = 40677
    MUL_ANONYM_MFEAT_FACTORS = 12
    MUL_ANONYM_MICRO_MASS = 1515
    MUL_ANONYM_PENDIGITS = 32
    MUL_ANONYM_POL = 201
    MUL_ANONYM_ROBERT = 41165
    MUL_ANONYM_SPOKEN_ARABIC_DIGIT = 1503
    MUL_ANONYM_SYNTHETIC_CONTROL = 377
    MUL_ANONYM_VOLCANOES_VENUS = 1542
    MUL_ANONYM_VOLKERT = 41166
    MUL_ANONYM_ZERNIKE_MFEAT = 22
    MUL_COMPUTERS_IMAGE_ARTIFICIAL_CHARACTERS = 1459
    MUL_COMPUTERS_IMAGE_CIFAR10 = 40927
    MUL_COMPUTERS_IMAGE_GESTURE_PHASE_SEGMENTATION = 4538
    MUL_COMPUTERS_IMAGE_GTSRB_GERMAN_TRAFFIC_SIGN = 41990
    MUL_COMPUTERS_IMAGE_INDIAN_PINES = 41972
    MUL_COMPUTERS_IMAGE_LETTER_RECOGNITION = 6
    MUL_COMPUTERS_IMAGE_MNIST_DIGITS = 554
    MUL_COMPUTERS_IMAGE_MNIST_FASHION = 40996
    MUL_COMPUTERS_IMAGE_MNIST_JAPANESE_KUZUSHIJI_49 = 41991
    MUL_COMPUTERS_IMAGE_OPT_DIGITS = 28
    MUL_COMPUTERS_IMAGE_SEGMENTATION = 40984
    MUL_COMPUTERS_INTRUSION_DETECTION_KDD = 42746
    MUL_COMPUTERS_META_STREAM_INTERVALS = 279
    MUL_COMPUTERS_PAGE_BLOCK_PARSING = 30
    MUL_COMPUTERS_ROBOT_WALL_NAVIGATION = 1497
    MUL_CONSUMER_ELECTRICITY_TAMILNADU = 40985
    MUL_CONSUMER_INTERNET_USAGE_PROFESSION = 372
    MUL_CONSUMER_PRODUCT_SENTIMENT = 46651
    MUL_CONSUMER_WOMEN_ECOMMERCE_CLOTHING_REVIEW = 46659
    MUL_FINANCIAL_CORPORATE_CREDIT_RATING_CAT = 46372
    MUL_FINANCIAL_CORPORATE_CREDIT_RATING_SYMBOL = 43344
    MUL_FINANCIAL_CREDIT_RISK_CHINA = 46444
    MUL_FINANCIAL_CREDIT_SCORE_HZL = 46441
    MUL_FINANCIAL_LOAN_IBRD = 46467
    MUL_FINANCIAL_P2P_LOAN = 43502
    MUL_FOOD_WINE_QUALITY_CAT = 40498
    MUL_FOOD_WINE_REVIEW = 46653
    MUL_GENETICS_SPLICE_DNA = 46
    MUL_GENETICS_ACP_BREAST_CANCER = 42895
    MUL_HEALTHCARE_APPENDICITIS_REGENSBURG_PEDIATRIC = 46603
    MUL_HEALTHCARE_AUDIOLOGY_DIAGNOSTIC = 7
    MUL_HEALTHCARE_BABY_SLEEP_STATE = 679
    MUL_HEALTHCARE_CANCER_CERVICAL = 46592
    MUL_HEALTHCARE_CMC_CONTRACEPTIVE_WOMEN = 23
    MUL_HEALTHCARE_DIABETES_US130 = 4541
    MUL_HEALTHCARE_EGYPTIAN_SKULLS = 1099
    MUL_HEALTHCARE_HEART_ARRHYTMIA = 5
    MUL_HEALTHCARE_HEART_CARDIOTOCOGRAPHY = 1466
    MUL_HEALTHCARE_HEPATITIS_C_EGYPT = 46607
    MUL_HEALTHCARE_HYPOTHYROID = 57
    MUL_HEALTHCARE_LYMPH_YUGOSLAVIA = 10
    MUL_HEALTHCARE_MENTAL_HEALTH_OCCUPATION = 46721
    MUL_HEALTHCARE_OBESITY_LEVELS = 46597
    MUL_HEALTHCARE_PBCSEQ_BILIARY_CIRRHOSIS = 516
    MUL_HEALTHCARE_TEETH_DMFT_ANALCATDATA = 469
    MUL_HEALTHCARE_WHITE_BLOOD_CELLS_WBC = 46676
    MUL_HOUSES_MELBOURNE_AIRBNB = 46665
    MUL_INDUSTRIAL_STEEL_PLATES_FAULT = 40982
    MUL_NATURE_AIR_QUALITY_POLLUTION = 46762
    MUL_NATURE_CJS_MAPLE_TREES = 23380
    MUL_NATURE_EUCALYPTUS_SEED = 188
    MUL_NATURE_FOREST_COVERTYPE = 1596
    MUL_NATURE_SOLAR_FLARES = 44966
    MUL_NATURE_SOYBEAN_DISEASE = 42
    MUL_NATURE_YEAST_PROTEIN = 181
    MUL_PROFESSIONAL_DATA_SCIENTIST_SALARY = 46664
    MUL_PROFESSIONAL_NURSERY_APPLICATIONS_SLOVENIA = 26
    MUL_PROFESSIONAL_STUDENT_PERFORMANCE_ENTRANCE_EXAM = 46584
    MUL_PROFESSIONAL_TEACHER_ASSISTANT_EVALUATION_TAE = 48
    MUL_SCIENCE_ANNEAL_CHEMICAL = 2
    MUL_SCIENCE_BALANCE_SCALE = 11
    MUL_SCIENCE_BATCH_CHORAL_HARMONY = 4552
    MUL_SCIENCE_DRUG_DIRECTORY_FDA_UNAPPROVED = 43044
    MUL_SCIENCE_MICE_PROTEIN = 40966
    MUL_SCIENCE_MIMIC_ROUTE = 46678
    MUL_SCIENCE_QSAR_BIOCONCENTRATION = 46585
    MUL_SCIENCE_VOWEL_JAPANESE = 375
    MUL_SCIENCE_VOWEL_SPEAKER_RECOGNITION = 307
    MUL_SCIENCE_WAVEFORM_500 = 60
    MUL_SOCIAL_BIAS_FRAMES = 46701
    MUL_SOCIAL_BOOKS_AUTHORSHIP_ANALCATDATA = 458
    MUL_SOCIAL_BROWN_FROWN_CORPORA_COLLINS = 40971
    MUL_SOCIAL_DPBEDIA = 46686
    MUL_SOCIAL_FUNPEDIA = 46692
    MUL_SOCIAL_GOOGLE_QA_TYPE_REASON = 46658
    MUL_SOCIAL_HARRY_POTTER_FAN_FICTION = 43544
    MUL_SOCIAL_HOLISTIC_BIAS = 46684
    MUL_SOCIAL_NEWS_CHANNEL_CATEGORY = 46652
    MUL_SOCIAL_OKCUPID_DATING_JOB_STEM = 42734
    MUL_SOCIAL_SOCC_OPINION_COMMENTS_CORPUS_TOXICITY = 46709
    MUL_SOCIAL_STACKOVERFLOW_POLARITY = 43160
    MUL_SOCIAL_SUPREME_COURT_ANALCATDATA = 504
    MUL_SOCIAL_WIKIPEDIA_TALK_LABELS_ATTACKS = 46708
    MUL_SPORTS_ACTIVITY_LOCALIZED_LDPA = 1483
    MUL_SPORTS_BASEBALL_HALL_OF_FAME = 185
    MUL_SPORTS_CHESS_JUNGLE = 41027
    MUL_SPORTS_CONNECT4_GAME = 40668
    MUL_SPORTS_POKER_HAND = 1567
    MUL_SPORTS_WALKING_ACTIVITY = 1509
    MUL_TRANSPORTATION_AIRLINES_TWITTER_SENTIMENT = 43397
    MUL_TRANSPORTATION_CAR_ACCEPTABILITY = 40975
    MUL_TRANSPORTATION_CAR_ORIGIN_COUNTRY = 455
    MUL_TRANSPORTATION_CAR_STUDENT_INSURANCE_RISK = 43157
    MUL_TRANSPORTATION_SHUTTLE_SPACE = 40685
    MUL_TRANSPORTATION_TRAFFIC_ACCIDENTS_FARS = 40672
    MUL_TRANSPORTATION_TRAFFIC_VIOLATION = 42345
    MUL_TRANSPORTATION_VEHICLE_SILHOUETTE = 54
    REG_ANONYM_ALLSTATE_CLAIM_SEVERITY = 42571
    REG_ANONYM_BANK_32NH = 558
    REG_ANONYM_BUZZ_IN_SOCIAL_MEDIA_TWITTER = 4549
    REG_ANONYM_DIONIS = 41167
    REG_ANONYM_FAT_MEAT_TECATOR = 505
    REG_ANONYM_GEOGRAPHICAL_ORIGIN_OF_MUSIC = 44965
    REG_ANONYM_HOUSE_16H = 574
    REG_ANONYM_MERCEDES_BENZ_GREENER_MANUFACTURING = 42570
    REG_ANONYM_PLANTS_TEXTURE = 1493
    REG_ANONYM_SANTANDER_TRANSACTION_VALUE = 42572
    REG_ANONYM_TOPO = 422
    REG_ANONYM_YOLANDA = 42705
    REG_ANONYM_YPROP = 416
    REG_COMPUTERS_ALGO_RUNTIME_MIP2016 = 43071
    REG_COMPUTERS_ALGO_RUNTIME_SAT11 = 41980
    REG_COMPUTERS_CPU_ACTIVITY = 44978
    REG_COMPUTERS_GAMING_FRAMES_FPS_BENCHMARK = 44992
    REG_COMPUTERS_META_LEVEL_LEARNING = 566
    REG_COMPUTERS_PUMA_ROBOT_ARM = 44981
    REG_COMPUTERS_ROBOT_KIN8NM = 44980
    REG_COMPUTERS_ROBOT_SARCOS = 44976
    REG_COMPUTERS_YOUTUBE_VIDEO_TRANSCODING = 44974
    REG_CONSUMER_AMERICAN_EAGLE_PRICES = 46656
    REG_CONSUMER_AVOCADO_SALES = 41210
    REG_CONSUMER_BLACK_FRIDAY = 41540
    REG_CONSUMER_BOOK_PRICE_PREDICTION = 46663
    REG_CONSUMER_BOOKING_HOTEL_REVIEW = 43712
    REG_CONSUMER_DIAMONDS_PRICES = 42225
    REG_CONSUMER_JC_PENNEY_PRODUCT_PRICE = 46661
    REG_CONSUMER_MEDICAL_CHARGES = 44146
    REG_CONSUMER_MERCARI_ONLINE_MARKETPLACE = 46660
    REG_CONSUMER_ONLINE_NEWS_POPULARITY = 46662
    BIN_CONSUMER_REASONER_RECOMMENDATION = 46681
    REG_FINANCIAL_CC_USER_SEGMENTATION = 43618
    REG_FINANCIAL_INSURANCE_HEALTH_HOURS_WORKED_WIFE = 44993
    REG_FINANCIAL_INSURANCE_PREMIUM_DATA = 43463
    REG_FINANCIAL_STOCK_AEROSPACE = 223
    REG_FOOD_WINE_JUDGEMENT_SENSORY = 546
    REG_HEALTHCARE_BODY_FAT = 560
    REG_HEALTHCARE_DIABETES_SKLEARN = 44223
    REG_HEALTHCARE_POLLUTION_LA_MORTALITY_RMFTSA = 666
    REG_HOUSES_BOSTON_HOUSE = 531
    REG_HOUSES_BOSTON_PRICE_NOMINAL = 42563
    REG_HOUSES_BRAZILIAN_HOUSES = 42688
    REG_HOUSES_CALIFORNIA_HOUSES = 44977
    REG_HOUSES_CALIFORNIA_PRICES_2020 = 46669
    REG_HOUSES_GERMAN_HOUSES = 43342
    REG_HOUSES_HOUSEHOLD_MONTHLY_ELECTRICITY = 43588
    REG_HOUSES_IOWA_HOUSES_PRICES = 42165
    REG_HOUSES_LISBON = 43660
    REG_HOUSES_MIAMI = 44983
    REG_HOUSES_PERTH = 43822
    REG_HOUSES_SEATTLE_KINGS_COUNTY = 42731
    REG_NATURE_ABALONE_FISH_RINGS = 42726
    REG_NATURE_CLIMATE_CHANGE_AGRICULTURE_FINANCIAL = 46726
    REG_NATURE_EL_NINO_KDD = 563
    REG_NATURE_FISH_TOXICITY = 44970
    REG_NATURE_FOREST_FIRES = 44962
    REG_NATURE_MYANMAR_AIR_QUALITY = 43748
    REG_NATURE_NO2_POLLUTION_NORWAY = 547
    REG_NATURE_POLLEN_GRAINS_SYNTH = 529
    REG_NATURE_POLLEN_LUXEMBOURG = 43648
    REG_NATURE_QUAKE_RICHTER = 550
    REG_NATURE_WIND_SPEED_IRELAND = 503
    REG_PROFESSIONAL_CPS88_WAGES = 44984
    REG_PROFESSIONAL_EMPLOYEE_SALARY_MONTGOMERY = 42125
    REG_PROFESSIONAL_STUDENT_PERFORMANCE_PORTUGAL = 44967
    REG_SCIENCE_AIRFOIL_SELF_NOISE = 44957
    REG_SCIENCE_AUCTION_VERIFICATION = 44958
    REG_SCIENCE_CONCRETE_COMPRESSIVE_STRENGTH = 44959
    REG_SCIENCE_ENERGY_EFFICIENCY = 44960
    REG_SCIENCE_GRID_STABILITY = 44973
    REG_SCIENCE_PHYSIOCHEMICAL_PROTEIN = 44963
    REG_SCIENCE_QSAR_TID_11 = 3050
    REG_SCIENCE_QSAR_TID_10980 = 3277
    REG_SCIENCE_SIMULATION_2D_PLANES = 215
    REG_SCIENCE_SIMULATION_FRIED_FORMULA_PREDICT = 564
    REG_SCIENCE_SIMULATION_MV = 344
    REG_SCIENCE_SULFUR = 44145
    REG_SCIENCE_SUPERCONDUCTIVITY = 44964
    REG_SCIENCE_WAVE_ENERGY = 44975
    REG_SOCIAL_COLLEGES_GRANTS = 42727
    REG_SOCIAL_OCCUPATION_MOBILITY_SOCMOB = 541
    REG_SOCIAL_OKCUPID_DATING_PROFILE_AGE = 42164
    REG_SOCIAL_STRIKES_PER_COUNTRY = 549
    REG_SOCIAL_US_CRIME = 42730
    REG_SOCIAL_VOTING_SPACE_GEOGRAPHIC_ANALYSIS = 507
    REG_SPORTS_BASEBALL_HITTER_SALARY = 525
    REG_SPORTS_FIFA20_PLAYERS_VALUE = 43766
    REG_SPORTS_FIFA22_WAGES = 45012
    REG_SPORTS_FIFA_PLAYERS_STATS = 42194
    REG_SPORTS_MONEYBALL = 41021
    REG_SPORTS_NBA_2K20_PLAYERS_RATING = 43420
    REG_SPORTS_NBA_ALL_STAR = 43653
    REG_TRANSPORTATION_CAR_GM_PRICE = 44994
    REG_TRANSPORTATION_FLIGHT_SIMULATION_ELEVATORS = 216
    REG_TRANSPORTATION_NAVAL_PROPULSION_PLANT = 44969
    REG_TRANSPORTATION_NYC_BIKE_TRIP_DURATION = 43573
    REG_TRANSPORTATION_NYC_TAXI_TIP = 42729
    REG_TRANSPORTATION_NYC_TAXI_TRIP_DURATION = 43584
    REG_TRANSPORTATION_SEOUL_BIKE_SHARING_DEMAND = 46328
    REG_TRANSPORTATION_US_AIRPORT_PASSENGERS = 43479
    REG_TRANSPORTATION_US_BIKE_SHARING_DEMAND = 42712
    REG_TRANSPORTATION_ZURICH_PUBLIC_TRANSPORT_DELAY = 40753

class KaggleDatasetID(Enum):
    BIN_SOCIAL_HUMAN_CHOICE_PREDICTION_LM_GAMES = "eilamshapira/human-choice-prediction-in-language-based-games/OPE_train.csv"
    MUL_FOOD_MICHELIN_GUIDE_RESTAURANTS = "ngshiheng/michelin-guide-restaurants-2021/michelin_my_maps.csv"
    MUL_FOOD_YELP_REVIEWS = "omkarsabnis/yelp-reviews-dataset/yelp.csv"
    MUL_SOCIAL_GOT_SCRIPTS = 'albenft/game-of-thrones-script-all-seasons/Game_of_Thrones_Script.csv'
    MUL_SOCIAL_US_ELECTIONS_SPEECHES = "imuhammad/us-2020-presidential-election-speeches/us_2020_election_speeches.csv"
    MUL_TRANSPORTATION_US_ACCIDENTS_MARCH23 = "sobhanmoosavi/us-accidents/US_Accidents_March23.csv"
    REG_CONSUMER_CAR_PRICE_CARDEKHO = "sukritchatterjee/used-cars-dataset-cardekho/cars_details_merges.csv"
    REG_FOOD_ALCOHOL_WIKILIQ_PRICES = "limtis/wikiliq-dataset/spirits_data.csv"
    REG_FOOD_BEER_RATINGS = "ruthgn/beer-profile-and-ratings-data-set/beer_profile_and_ratings.csv"
    REG_FOOD_CHOCOLATE_BAR_RATINGS = "rtatman/chocolate-bar-ratings/flavors_of_cacao.csv"
    REG_FOOD_COFFEE_REVIEW = "hanifalirsyad/coffee-scrap-coffeereview/coffee_clean.csv"
    REG_FOOD_RAMEN_RATINGS_2022 = "ankanhore545/top-ramen-ratings-2022/Top Ramen Ratings .csv"
    REG_FOOD_WHISKY_SCOTCH_REVIEWS = "neilcosgrove/scotch-whiskey-reviews-update-2020/scotch_review2020.csv"
    REG_FOOD_WINE_POLISH_MARKET_PRICES = "skamlo/wine-price-on-polish-market/wine.csv"
    REG_FOOD_WINE_VIVINO_SPAIN = 'joshuakalobbowles/vivino-wine-data/vivino.csv'
    REG_FOOD_ZOMATO_RESTAURANTS = "himanshupoddar/zomato-bangalore-restaurants/zomato.csv"
    REG_PROFESSIONAL_COMPANY_EMPLOYEES_SIZE = "peopledatalabssf/free-7-million-company-dataset/companies_sorted.csv"
    REG_SOCIAL_ANIME_PLANET_RATING = 'hernan4444/animeplanet-recommendation-database-2020/anime.csv'
    REG_SOCIAL_BOOK_READABILITY_CLEAR = "verracodeguacas/clear-corpus/CLEAR.csv"
    REG_SOCIAL_FILMTV_MOVIE_RATING_ITALY = "stefanoleone992/filmtv-movies-dataset/filmtv_movies.csv"
    REG_SOCIAL_KOREAN_DRAMA = "noorrizki/top-korean-drama-list-1500/kdrama_list.csv"
    REG_SOCIAL_MOVIES_DATASET_REVENUE = "rounakbanik/the-movies-dataset/movies_metadata.csv"
    REG_SOCIAL_MUSEUMS_US_REVENUES = "markusschmitz/museums/museums_prep.csv"
    REG_SOCIAL_SPOTIFY_POPULARITY = "maharshipandya/-spotify-tracks-dataset/dataset.csv"
    REG_SOCIAL_VIDEO_GAMES_SALES = "gregorut/videogamesales/vgsales.csv"
    REG_SPORTS_FOOTBALL_MANAGER_IMPORTANT_MATCHES = "ajinkyablaze/football-manager-data/dataset.csv"
    REG_SPORTS_NBA_DRAFT_VALUE_OVER_REPLACEMENT = "mattop/nba-draft-basketball-player-data-19892021/nbaplayersdraft.csv"
    REG_TRANSPORTATION_USED_CAR_MERCEDES_BENZ_ITALY = "bogdansorin/second-hand-mercedes-benz-registered-2000-2023-ita/mercedes-benz.csv"
    REG_TRANSPORTATION_USED_CAR_PAKISTAN = "mustafaimam/used-car-prices-in-pakistan-2021/Used_car_prices_in_Pakistan_cleaned.csv"
    REG_TRANSPORTATION_USED_CAR_SAUDI_ARABIA = "turkibintalib/saudi-arabia-used-cars-dataset/UsedCarsSA_Clean_EN.csv"

class UrlDatasetID(Enum):
    REG_CONSUMER_BABIES_R_US_PRICES = "http://pages.cs.wisc.edu/~anhai/data/784_data/baby_products/csv_files/babies_r_us.csv"
    REG_CONSUMER_BIKE_PRICE_BIKEWALE = "http://pages.cs.wisc.edu/~anhai/data/784_data/bikes/csv_files/bikewale.csv"
    REG_PROFESSIONAL_EMPLOYEE_RENUMERATION_VANCOUBER = "https://opendata.vancouver.ca/api/records/1.0/download/?dataset=employee-remuneration-and-expenses-earning-over-75000&format=csv"
    REG_PROFESSIONAL_ML_DS_AI_JOBS_SALARIES = "https://ai-jobs.net/salaries/download/salaries.csv"
    REG_PROFESSIONAL_SCIMAGOJR_ACADEMIC_IMPACT = "https://www.scimagojr.com/journalrank.php?out=xls"
    REG_SOCIAL_BOOKS_GOODREADS = "http://pages.cs.wisc.edu/~anhai/data/784_data/books2/csv_files/goodreads.csv"
    REG_SOCIAL_MOVIES_ROTTEN_TOMATOES = "http://pages.cs.wisc.edu/~anhai/data/784_data/movies1/csv_files/rotten_tomatoes.csv"


TabularDatasetID = Union[OpenMLDatasetID, KaggleDatasetID, UrlDatasetID]


from typing import Union

import numpy as np
from pandas import Series
from sklearn.preprocessing import LabelEncoder, StandardScaler

from tabstar.preprocessing.scaler import fit_standard_scaler, transform_clipped_z_scores


def transform_preprocess_y(y: Union[Series, np.ndarray], scaler: Union[LabelEncoder, StandardScaler]) -> Series:
    y = y.copy()
    if isinstance(scaler, StandardScaler):
        return transform_clipped_z_scores(s=y, scaler=scaler)
    elif isinstance(scaler, LabelEncoder):
        return transform_cls_y(y=y, encoder=scaler)
    raise TypeError(f"What is this scaler {scaler} from type {type(scaler)}?")


def transform_cls_y(y: Series, encoder: LabelEncoder) -> Series:
    y_val = encoder.transform(y)
    return Series(y_val, name=y.name, index=y.index)


def fit_preprocess_y(y: Series, is_cls: bool) -> Union[LabelEncoder, StandardScaler]:
    if is_cls:
        return fit_cls_y(y)
    else:
        return fit_reg_y(y)

def fit_cls_y(y: Series) -> LabelEncoder:
    label_encoder = LabelEncoder()
    label_encoder.fit(y)
    return label_encoder

def fit_reg_y(y: Series) -> StandardScaler:
    return fit_standard_scaler(s=y)


import gc
import shutil
import time
from typing import Tuple, Optional

import numpy as np
import torch
from peft import PeftModel
from torch import Tensor
from torch.amp import autocast, GradScaler
from torch.utils.data import DataLoader
from tqdm import tqdm

from tabstar.tabstar_verbalizer import TabSTARData
from tabstar.training.checkpoint_averaging import CheckpointManager
from tabstar.training.dataloader import get_dataloader
from tabstar.training.early_stopping import EarlyStopping
from tabstar.training.hyperparams import set_accumulation_steps
from tabstar.training.lora import load_pretrained, load_finetuned
from tabstar.training.metrics import calculate_metric, apply_loss_fn, calculate_loss
from tabstar.training.optimizer import get_optimizer, get_scheduler
from tabstar.training.utils import concat_predictions


class TabStarTrainer:

    def __init__(self, max_epochs: int, lora_lr: float, lora_r: int, lora_batch: int, patience: int,
                 global_batch: int, device: torch.device, model_version: str, cp_average: bool, time_limit: int,
                 output_dir: Optional[str], metric_name: Optional[str]):
        self.lora_lr = lora_lr
        self.lora_batch = lora_batch
        self.global_batch = global_batch
        self.accumulation_steps = set_accumulation_steps(global_batch=global_batch, batch_size=lora_batch)
        self.max_epochs = max_epochs
        self.device = device
        self.cp_average = cp_average
        self.model_version = model_version
        self.metric_name = metric_name
        self.model = load_pretrained(model_version=model_version, lora_r=lora_r)
        self.model.to(self.device)
        self.optimizer = get_optimizer(model=self.model, lr=self.lora_lr)
        self.scheduler = get_scheduler(optimizer=self.optimizer, max_lr=self.lora_lr, epochs=self.max_epochs)
        self.use_amp = bool(self.device.type == "cuda")
        self.scaler = GradScaler(enabled=self.use_amp)
        self.early_stopper = EarlyStopping(patience=patience, metric_name=metric_name)
        self.cp_manager = CheckpointManager(do_average=self.cp_average, output_dir=output_dir)
        self.steps: int = 0
        self.time_limit = time_limit or 60 * 60 * 10

    def train(self, train_data: TabSTARData, val_data: TabSTARData) -> float:
        train_loader = get_dataloader(train_data, is_train=True, batch_size=self.lora_batch)
        val_loader = get_dataloader(val_data, is_train=False)
        start_time = time.time()
        for epoch in tqdm(range(1, self.max_epochs + 1), desc="Epochs", leave=False):
            train_loss = self._train_epoch(train_loader)
            val_loss, val_metric = self._evaluate_epoch(val_loader)
            emoji = self.early_stopper.update_metric(metric=val_metric)
            print(f"Epoch {epoch} || Train {train_loss:.4f} || Val {val_loss:.4f} || Metric {val_metric:.4f} {emoji}")
            if self.early_stopper.is_best:
                self.model.save_pretrained(self.cp_manager.best_dir)
            elif self.early_stopper.should_stop:
                print(f"🛑 Early stopping at epoch {epoch}")
                break
            self.scheduler.step()
            self.cp_manager.save_checkpoint(model=self.model, epoch=epoch, val_loss=val_loss)
            if self.will_next_epoch_exceed_budget(epoch=epoch, start_time=start_time):
                break
        self.cp_manager.average_checkpoints(model=self.model, evaluator=self._evaluate_epoch, val_loader=val_loader)
        best_metric = self.cp_manager.avg_metric or self.early_stopper.metric
        return best_metric

    def _train_epoch(self, dataloader: DataLoader) -> float:
        self.model.train()
        total_loss = 0.0
        total_samples = 0
        for data in tqdm(dataloader, desc="Batches", leave=False):
            batch_loss = self._train_batch(data)
            total_loss += batch_loss * len(data.y)
            total_samples += len(data.y)
            self.steps += 1
            if self.steps % self.accumulation_steps == 0:
                self._do_update()
        if self.steps % self.accumulation_steps != 0:
            self._do_update()
        epoch_loss = total_loss / total_samples
        return epoch_loss

    def _train_batch(self, data: TabSTARData) -> float:
        with autocast(device_type=self.device.type, enabled=self.use_amp):
            loss, predictions = self._do_forward(data=data)
            loss_for_backward = loss / self.accumulation_steps
        if self.use_amp:
            scaled_loss = self.scaler.scale(loss_for_backward)
            scaled_loss.backward()
        else:
            loss_for_backward.backward()
        original_mean_batch_loss = loss.item()
        return original_mean_batch_loss

    def _do_forward(self, data: TabSTARData) -> Tuple[Tensor, Tensor]:
        predictions = self.model(x_txt=data.x_txt, x_num=data.x_num, d_output=data.d_output)
        loss = calculate_loss(predictions=predictions, y=data.y, d_output=data.d_output)
        return loss, predictions

    def _do_update(self):
        self.scaler.unscale_(self.optimizer)
        torch.nn.utils.clip_grad_norm_(self.model.parameters(), max_norm=1.0)
        self.scaler.step(self.optimizer)
        self.scaler.update()
        self.optimizer.zero_grad()

    def _evaluate_epoch(self, dataloader: DataLoader) -> Tuple[float, float]:
        self.model.eval()
        total_loss = 0.0
        total_samples = 0

        y_pred = []
        y_true = []
        d_output = None

        for data in dataloader:
            d_output = data.d_output
            with torch.no_grad(), autocast(device_type=self.device.type, enabled=self.use_amp):
                batch_loss, batch_predictions = self._do_forward(data=data)
                total_loss += batch_loss * len(data.y)
                total_samples += len(data.y)
                batch_predictions = apply_loss_fn(prediction=batch_predictions, d_output=d_output)
                y_pred.append(batch_predictions)
                y_true.append(data.y)
        y_pred = concat_predictions(y_pred)
        y_true = np.concatenate(y_true)
        metrics = calculate_metric(y_true=y_true, y_pred=y_pred, d_output=d_output, metric_name=self.metric_name)
        loss = total_loss / total_samples
        loss = loss.item()
        return loss, metrics.score

    def load_model(self) -> PeftModel:
        self.model = None
        if torch.cuda.is_available():
            torch.cuda.empty_cache()
        gc.collect()
        self.model = load_finetuned(self.cp_manager.to_load_dir, tabstar_version=self.model_version)
        self.model.to(self.device)
        self.model.eval()
        return self.model

    def delete_model(self):
        shutil.rmtree(self.cp_manager.to_load_dir)

    def will_next_epoch_exceed_budget(self, epoch: int, start_time) -> bool:
        elapsed = round(time.time() - start_time, 1)
        avg_epoch_time = round(elapsed / epoch, 1)
        next_epoch_estimate = round(elapsed + avg_epoch_time, 1)
        if next_epoch_estimate > self.time_limit:
            print(f"⏱️ Limit exceeds next epoch: {elapsed=}, {avg_epoch_time=}, {self.time_limit=}. Stopping!")
            return True
        return False



import torch
from pandas import Series

from tabstar.training.metrics import calculate_loss


def test_regression_loss():
    predictions = torch.tensor([[2.5], [0.0], [2.1]])
    y = Series([3.0, -0.5, 2.0])
    loss = calculate_loss(predictions, y, d_output=1)
    assert round(loss.item(), 4) == 0.17

def test_multiclass_loss():
    predictions = torch.tensor([[1.0, 2.0, 0.5, -1.0], [2.0, 0.5, 1.0, -0.5]])
    y = Series([1, 2])
    loss = calculate_loss(predictions, y, d_output=10)
    assert round(loss.item(), 4) == 1.0049

def test_binary_class_loss():
    predictions = torch.tensor([[2.0, 1.0], [0.5, 2.5]])
    y = Series([0, 1])
    loss = calculate_loss(predictions, y, d_output=2)
    assert round(loss.item(), 4) == 0.2201


import torch
import torch.nn as nn

from tabstar.arch.config import D_MODEL


class PredictionHead(nn.Module):

    def __init__(self, input_size: int = D_MODEL):
        super().__init__()
        hidden_size = input_size * 4
        self.layers = nn.Sequential(
            nn.Linear(input_size, hidden_size),
            nn.ReLU(),
            nn.Linear(hidden_size, 1)
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.layers(x)



import re
from typing import Tuple, Optional

from pandas import DataFrame, Series


def replace_column_names(x: DataFrame, y: Optional[Series]) -> Tuple[DataFrame, Optional[Series]]:
    old2new = {c: normalize_col_name(c) for c in x.columns}
    x.rename(columns=old2new, inplace=True)
    if y is not None:
        y.name = normalize_col_name(str(y.name))
    return x, y

def normalize_col_name(text: str) -> str:
    for c in ['\n', '\r', '\t', '\u00A0']:
        text = text.replace(c, ' ')
    text = replace_unspaced_symbols(text)
    text = replace_whitespaces(text)
    return text

def replace_unspaced_symbols(text: str) -> str:
    if ' ' not in text:
        return text
    for c in ['_', '-', ".", ":"]:
        text = text.replace(c, ' ')
    return text

def replace_whitespaces(text: str) -> str:
    # TODO: originally this was supposed to keep newlines and tabs
    return re.sub(r'[\x00-\x1F\x7F]', ' ', text)



from typing import Type

import pytest
import torch

from tabstar.datasets.all_datasets import OpenMLDatasetID
from tabstar_paper.baselines.abstract_model import TabularModel
from tabstar_paper.baselines.catboost import CatBoost
from tabstar_paper.benchmarks.evaluate import evaluate_on_dataset


def _test_evaluate(model_cls: Type[TabularModel], fold: int = 0) -> float:
    d_summary = evaluate_on_dataset(model_cls=model_cls, dataset_id=OpenMLDatasetID.BIN_FINANCIAL_CREDIT_GERMAN,
                                    fold=fold, device=torch.device("cpu"), train_examples=100, verbose=False)
    return d_summary["test_score"]




def test_catboost():
    f0 = _test_evaluate(model_cls=CatBoost, fold=0)
    f1 = _test_evaluate(model_cls=CatBoost, fold=0)
    expected_score = 0.6595
    assert f0 == f1 == pytest.approx(expected_score, abs=1e-4)
    f0 = _test_evaluate(model_cls=CatBoost, fold=1)
    f1 = _test_evaluate(model_cls=CatBoost, fold=1)
    expected_score = 0.6852
    assert f0 == f1 == pytest.approx(expected_score, abs=1e-4)



from dataclasses import dataclass
from typing import Union, Dict, Optional

import numpy as np
import torch
from numpy.exceptions import AxisError
from pandas import Series
from sklearn.metrics import roc_auc_score, r2_score, mean_squared_error, accuracy_score, f1_score, log_loss, recall_score
from torch import Tensor, softmax
from torch.nn import CrossEntropyLoss, MSELoss


LOGLOSS = "logloss"
RMSE = "rmse"
R2 = "r2"
ROC_AUC = "roc_auc"
ROC_AUC_OVR = "roc_auc_ovr"

METRICS_TO_MINIMIZE = [LOGLOSS, RMSE]

@dataclass
class Metrics:
    score: float
    metrics: Dict[str, float]

    def __post_init__(self):
        self.score = float(self.score)
        self.metrics = {k: float(v) for k, v in self.metrics.items()}


def calculate_metric(y_true: Union[np.ndarray, Series], y_pred: np.ndarray, d_output: int, is_pretrain: bool = False,
                     metric_name: Optional[None] = None) -> Metrics:
    if d_output == 1:
        return _calculate_metrics_for_regression(y_true=y_true, y_pred=y_pred, is_pretrain=is_pretrain,
                                                 metric_name=metric_name)
    elif d_output == 2:
        if y_pred.ndim == 2 and y_pred.shape[1] == 2:
            y_pred = y_pred[:, 1]
        return _calculate_metrics_for_binary(y_true=y_true, y_pred=y_pred, metric_name=metric_name)
    elif d_output > 2:
        return _calculate_metrics_for_multiclass(y_true=y_true, y_pred=y_pred, metric_name=metric_name)
    raise ValueError(f"Unsupported d_output: {d_output}. Expected 1 (regression), 2 (binary), or >2 (multiclass).")

def _calculate_metrics_for_regression(y_true: Union[np.ndarray, Series], y_pred: np.ndarray, is_pretrain: bool,
                                      metric_name: Optional[str]) -> Metrics:
    rsq = r2_score(y_true=y_true, y_pred=y_pred)
    mse = mean_squared_error(y_true=y_true, y_pred=y_pred)
    rmse = np.sqrt(mse)
    one_minus_mse = 1 - mse
    metrics = {R2: rsq, 'mse': mse, '1-mse': one_minus_mse, RMSE: rmse}
    if is_pretrain:
        score = one_minus_mse
    elif metric_name == RMSE:
        score = rmse
    else:
        if (metric_name is not None) and metric_name not in [RMSE, R2]:
            print(f"⚠️ Unsupported metric_name for regression! {metric_name}. We want: {RMSE} or {R2}.")
        score = rsq
    return Metrics(score=score, metrics=metrics)

def _calculate_metrics_for_binary(y_true: Union[np.ndarray, Series], y_pred: np.ndarray, metric_name: Optional[str]) -> Metrics:
    y_pred_label = (y_pred > 0.5).astype(int)
    auc = roc_auc_score(y_true, y_pred)
    acc = accuracy_score(y_true, y_pred_label)
    f1 = f1_score(y_true, y_pred_label)
    logloss = log_loss(y_true, y_pred)
    recall = recall_score(y_true, y_pred_label)
    metrics = {ROC_AUC: auc, LOGLOSS: logloss, 'accuracy': acc, 'f1': f1, 'recall': recall}
    if metric_name is not None and metric_name != ROC_AUC:
        print(f"⚠️ Unsupported metric name: {metric_name}. We want: {ROC_AUC}")
    return Metrics(score=auc, metrics=metrics)

def _calculate_metrics_for_multiclass(y_true: Union[np.ndarray, Series], y_pred: np.ndarray, metric_name: Optional[str]) -> Metrics:
    try:
        auc = roc_auc_score(y_true=y_true, y_score=y_pred, multi_class='ovr', average='macro')
    except (ValueError, AxisError):
        # Error calculating AUC, likely due to class imbalance or insufficient samples
        auc = _per_class_auc(y_true=y_true, y_pred=y_pred)
    try:
        logloss = log_loss(y_true, y_pred)
    except Exception:
        logloss = np.nan
    y_pred_label = np.argmax(y_pred, axis=1)
    f1 = f1_score(y_true, y_pred_label, average='macro')
    acc = accuracy_score(y_true, y_pred_label)
    metrics = {
        ROC_AUC_OVR: auc,
        LOGLOSS: logloss,
        'macro_f1': f1,
        'accuracy': acc,
    }
    if metric_name == LOGLOSS:
        score = logloss
    else:
        if metric_name is not None and metric_name not in [ROC_AUC_OVR, LOGLOSS]:
            print(f"⚠️ Unsupported metric_name: {metric_name}. We want: {LOGLOSS} or {ROC_AUC_OVR}.")
        score = auc
    return Metrics(score=score, metrics=metrics)

def _per_class_auc(y_true, y_pred) -> float:
    present_classes = np.unique(y_true)
    aucs = {}
    for cls in present_classes:
        # Binary ground truth: 1 for the current class, 0 for others
        y_true_binary = (y_true == cls).astype(int)
        # Predicted probabilities for the current class
        y_pred_scores = y_pred[:, int(cls)]
        try:
            auc = roc_auc_score(y_true_binary, y_pred_scores)
            aucs[cls] = auc
        except ValueError:
            pass
    macro_avg = float(np.mean(list(aucs.values())))
    return macro_avg


def apply_loss_fn(prediction: Tensor, d_output: int) -> Tensor:
    if d_output == 1:
        return prediction.to(torch.float32)
    prediction = prediction.to(torch.float32)
    prediction = softmax(prediction, dim=1)
    if d_output == 2:
        # We want the probability of '1'
        prediction = prediction[:, 1]
    return prediction


def calculate_loss(predictions: Tensor, y: Union[Series, np.ndarray], d_output: int) -> Tensor:
    is_reg = bool(d_output == 1)
    if is_reg:
        loss_fn = MSELoss()
        dtype = torch.float32
    else:
        loss_fn = CrossEntropyLoss()
        dtype = torch.long
    if not isinstance(y, Tensor):
        y = torch.tensor(y, dtype=dtype)
    y = y.to(predictions.device)
    if is_reg and y.ndim == 1:
        y = y.unsqueeze(1)
    loss = loss_fn(predictions, y)
    return loss


MAX_EPOCHS = 50
FINETUNE_PATIENCE = 5
LORA_LR = 0.001
LORA_R = 32
LORA_BATCH = 64
GLOBAL_BATCH = 128


def set_accumulation_steps(batch_size: int, global_batch: int):
    if global_batch % batch_size != 0:
        raise ValueError("Global batch size must be divisible by local batch size.")
    if global_batch < batch_size:
        raise ValueError("Global batch size must be greater than or equal to local batch size.")
    return global_batch // batch_size


import os
from subprocess import Popen, PIPE
from typing import Optional, Union

import torch


CPU_CORES = 8

EXISTING_CORES = os.cpu_count()
if EXISTING_CORES < CPU_CORES:
    print(f"❗ Warning: {CPU_CORES} CPU devices requested, but only {EXISTING_CORES} available.")
    CPU_CORES = EXISTING_CORES

def get_device(device: Optional[Union[str, torch.device]] = None) -> torch.device:
    if isinstance(device, torch.device):
        return device
    gpu = os.getenv('CUDA_VISIBLE_DEVICES')
    if gpu is not None:
        print(f"🖥️ Using CUDA_VISIBLE_DEVICES={gpu}. Setting device to 'cuda'.")
        if isinstance(gpu, int) or gpu.isdigit():
            return torch.device("cuda")
        raise ValueError(f"Invalid CUDA_VISIBLE_DEVICES value: {gpu}. It should be an integer for now.")
    if device is None:
        device = _get_device_type()
    return torch.device(device)

def clear_cuda_cache():
    if torch.cuda.is_available():
        try:
            torch.cuda.empty_cache()
        except RuntimeError:
            pass

def _get_device_type() -> str:
    if torch.cuda.is_available():
        clear_cuda_cache()
        return _get_free_gpu()
    elif torch.backends.mps.is_available():
        torch.mps.empty_cache()
        return "mps"
    print(f"⚠️ No GPU available, using CPU. This may lead to slow performance.")
    return "cpu"

def _get_free_gpu() -> str:
    gpu_output = Popen(["nvidia-smi", "-q", "-d", "PIDS"], stdout=PIPE, encoding="utf-8")
    gpu_processes = Popen(["grep", "Processes"], stdin=gpu_output.stdout, stdout=PIPE, encoding="utf-8")
    gpu_output.stdout.close()
    processes_output = gpu_processes.communicate()[0]
    for i, line in enumerate(processes_output.strip().split("\n")):
        if line.endswith("None"):
            return f"cuda:{i}"
    raise RuntimeError("No free GPU found! However, pass `CUDA_VISIBLE_DEVICES` if you insist on using one.")


def get_gpu_num(device: str) -> int:
    prefix = "cuda:"
    assert device.startswith(prefix), f"Device {device} should start with {prefix}"
    return int(device.replace(prefix, ''))


from tabstar_paper.pretraining.unfreezing import get_last_layers_num


def test_unfreeze_layers():
    assert get_last_layers_num(total_layers=10, to_unfreeze=3) == [9, 8, 7]
    assert get_last_layers_num(total_layers=10, to_unfreeze=0) == []
    assert get_last_layers_num(total_layers=12, to_unfreeze=2) == [11, 10]
    assert get_last_layers_num(total_layers=12, to_unfreeze=12) == [11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0]


import random
from typing import Tuple, Optional

import numpy as np
import pandas as pd
from pandas import DataFrame, Series
from sklearn.model_selection import train_test_split

from tabstar.constants import SEED

TEST_RATIO = 0.1
MAX_TEST_SIZE = 2000
VAL_RATIO = 0.1
MAX_VAL_SIZE = 1000

def split_to_test(x: DataFrame, y: Series, is_cls: bool, fold: int = -1, train_examples: Optional[int] = None) -> Tuple[DataFrame, DataFrame, Series, Series]:
    test_size = int(len(y) * TEST_RATIO)
    test_size = min(test_size, MAX_TEST_SIZE)
    if (train_examples is not None) and len(x) > train_examples:
        test_size = len(x) - train_examples
    x_train, x_test, y_train, y_test = do_split(x=x, y=y, test_size=test_size, is_cls=is_cls, fold=fold)
    return x_train, x_test, y_train, y_test

def split_to_val(x: DataFrame, y: Series, is_cls: bool, fold: int = -1, val_ratio: float = VAL_RATIO) -> Tuple[DataFrame, DataFrame, Series, Series]:
    val_size = int(len(y) * val_ratio)
    val_size = min(val_size, MAX_VAL_SIZE)
    x_train, x_val, y_train, y_val = do_split(x=x, y=y, test_size=val_size, is_cls=is_cls, fold=fold)
    return x_train, x_val, y_train, y_val


def do_split(x: DataFrame, y: Series, test_size: int, is_cls: bool, fold: int) -> Tuple[DataFrame, DataFrame, Series, Series]:
    if not x.index.equals(y.index):
        raise ValueError("X and y must share identical indices. Construct y as: y = pd.Series(y, index=X.index)")
    random_state = SEED + fold
    if not is_cls:
        return train_test_split(x, y, test_size=test_size, random_state=random_state)
    num_classes = y.nunique()
    if num_classes == 2:
        return _split_for_binary_classification(x, y, test_size=test_size, random_state=random_state)
    has_rare_class = y.value_counts().min() <= 1
    if has_rare_class:
        return _split_with_rare_classes(x, y, test_size=test_size, random_state=random_state)
    return train_test_split(x, y, test_size=test_size, random_state=random_state, stratify=y)


def _split_with_rare_classes(x: DataFrame, y: Series, test_size: int, random_state: int) -> Tuple[DataFrame, DataFrame, Series, Series]:
    # TODO: add tests here, seems like a complex function
    singleton_classes = y.value_counts()[y.value_counts() == 1].index
    is_singleton = y.isin(singleton_classes)
    x_single = x[is_singleton]
    y_single = y[is_singleton]
    x_rest = x[~is_singleton]
    y_rest = y[~is_singleton]

    rest_classes = len(set(y_rest))
    test_size = max(test_size, rest_classes)
    x_train, x_test, y_train, y_test = train_test_split(x_rest, y_rest, test_size=test_size, random_state=random_state,
                                                        stratify=y_rest)
    x_train, y_train = merge_splits(x1=x_train, x2=x_single, y1=y_train, y2=y_single, random_state=random_state)
    return x_train, x_test, y_train, y_test


def _split_for_binary_classification(x: DataFrame, y: Series, test_size: int, random_state: int) -> Tuple[DataFrame, DataFrame, Series, Series]:
    test_indices = []
    y_values = sorted(set(y))
    assert len(y_values) == 2, f"Expected 2 classes, got {y_values}"
    random_number_generator = np.random.RandomState(random_state)
    for v in y_values:
        v_indices = y[y == v].index.tolist()
        v_idx = random_number_generator.choice(v_indices)
        test_indices.append(v_idx)
    x_single_test = x.loc[test_indices]
    y_single_test = y.loc[test_indices]
    x_rest = x.drop(index=test_indices)
    y_rest = y.drop(index=test_indices)
    x_train, x_test, y_train, y_test = train_test_split(x_rest, y_rest, test_size=test_size, random_state=random_state,
                                                        stratify=y_rest)
    x_test, y_test = merge_splits(x1=x_test, x2=x_single_test, y1=y_test, y2=y_single_test, random_state=random_state)
    return x_train, x_test, y_train, y_test



def merge_splits(x1: DataFrame, x2: DataFrame, y1: Series, y2: Series, random_state: int) -> Tuple[DataFrame, Series]:
    x_train = pd.concat([x1, x2])
    y_train = pd.concat([y1, y2])
    shuffled = x_train.sample(frac=1, random_state=random_state).index
    x = x_train.loc[shuffled]
    y = y_train.loc[shuffled]
    return x, y


from typing import Set, Optional

import pandas as pd
from pandas import DataFrame, Series
from pandas.core.dtypes.common import is_datetime64_any_dtype, is_numeric_dtype, is_object_dtype, is_bool_dtype

from tabstar.preprocessing.detection import is_mostly_numerical, is_numeric
from tabstar.preprocessing.nulls import convert_numeric_with_missing, MISSING_VALUE, get_valid_values

pd.set_option('future.no_silent_downcasting', True)


def transform_feature_types(x: DataFrame, numerical_features: Set[str]) -> DataFrame:
    new_x = {}
    for col in x.columns:
        if col in numerical_features:
            new_x[col] = convert_series_to_numeric(s=x[col])
        else:
            new_x[col] = convert_series_to_textual(s=x[col])
    new_x = pd.DataFrame(new_x, index=x.index)
    ordered_x = new_x[x.columns]
    return ordered_x


def detect_numerical_features(x: DataFrame) -> Set[str]:
    return {col for col in x.columns if is_numerical_feature(s=x[col])}


# TODO: for future versions, maybe best to rely on maintained packages, e.g. skrub's TableVectorizer
def is_numerical_feature(s: Series) -> bool:
    if is_datetime64_any_dtype(s.dtype):
        raise TypeError(f"At this point, dates should have already been transformed.")
    elif len(get_valid_values(s)) == 0:
        return False
    elif is_numeric_dtype(s.dtype) or is_mostly_numerical(s=s):
        return True
    elif is_object_dtype(s.dtype) or is_bool_dtype(s.dtype) or isinstance(s.dtype, pd.CategoricalDtype):
        return False
    else:
        raise ValueError(f"Unsupported dtype {s.dtype} for series {s.name}")


def convert_series_to_numeric(s: Series, missing_value: Optional[str] = None) -> Series:
    if pd.api.types.is_numeric_dtype(s):
        return s.astype(float)
    non_numeric_indices = [not is_numeric(f) for f in s]
    if not any(non_numeric_indices):
        return s.astype(float)
    if missing_value is None:
        unique_non_numeric = s[non_numeric_indices].unique()
        if len(unique_non_numeric) != 1:
            raise ValueError(f"Missing values detected are {unique_non_numeric}. Should be only one!")
        missing_value = unique_non_numeric[0]
    return convert_numeric_with_missing(s=s, missing_value=missing_value)


def convert_series_to_textual(s: Series):
    return s.astype(object).fillna(MISSING_VALUE).astype(str)


from typing import List, Optional

import pandas as pd
from pandas import DataFrame
from pandas.core.dtypes.common import is_numeric_dtype

from tabstar.preprocessing.texts import replace_whitespaces


def verbalize_textual_features(x: DataFrame) -> DataFrame:
    semantic_cols = [col for col, dtype in x.dtypes.items() if not is_numeric_dtype(dtype)]
    for col in semantic_cols:
        x[col] = x[col].apply(lambda v: verbalize_feature(col=str(col), value=v))
    return x

def verbalize_feature(col: str, value: str) -> str:
    if not isinstance(value, str):
        raise TypeError(f"Expected string value, got {type(value)}")
    col = replace_whitespaces(col)
    value = replace_whitespaces(value)
    v = f"Predictive Feature: {col}\nFeature Value: {value}"
    return v

def prepend_target_tokens(x: DataFrame, y_name: str, y_values: Optional[List[str]]) -> DataFrame:
    y_name = replace_whitespaces(y_name)
    if y_values:
        values = [replace_whitespaces(text=str(v)) for v in y_values]
        tokens = [f"Target Feature: {y_name}\nFeature Value: {v}" for v in values]
    else:
        tokens = [f"Numerical Target Feature: {y_name}"]
    target_df = DataFrame({f"TABSTAR_TARGET_TOKEN_{i}": [t] * len(x) for i, t in enumerate(tokens)}, index=x.index)
    x = pd.concat([target_df, x], axis=1)
    return x



from tabstar.preprocessing.binning import format_float


def test_numerical_verbalizations():
    assert format_float(3.1415926) == "3.1416"
    assert format_float(0.0000001) == "0"
    assert format_float(0.00001) == "0"
    assert format_float(0.0500) == "0.05"


import os
import shutil

import numpy as np

from tabstar.datasets.all_datasets import OpenMLDatasetID
from tabstar_paper.pretraining.dataloaders import get_dev_dataloader
from tabstar_paper.pretraining.datasets import create_pretrain_dataset


TEST_DATA_DIR = "temp_test_dir_imdb"

IMDB_X_TXT_FIRST_ROW = [
'Target Feature: Genre\nFeature Value: Drama',
'Target Feature: Genre\nFeature Value: Not Drama',
'Predictive Feature: Title\nFeature Value: Tall Men',
'Predictive Feature: Description\nFeature Value: A challenged man is stalked by tall phantoms in business suits after he purchases a car with a mysterious black credit card.',
'Predictive Feature: Director\nFeature Value: Jonathan Holbrook',
'Predictive Feature: Actors\nFeature Value: Dan Crisafulli, Kay Whitney, Richard Garcia, Pat Cashman',
'Predictive Feature: Metascore\nFeature Value: 55 to 59 (Quantile 40 - 50%)',
'Predictive Feature: Rank\nFeature Value: 598.4 to 691.3 (Quantile 60 - 70%)',
'Predictive Feature: Rating\nFeature Value: 1.9 to 5.59 (Quantile 0 - 10%)',
'Predictive Feature: Revenue (Millions)\nFeature Value: Unknown Value',
'Predictive Feature: Runtime (Minutes)\nFeature Value: 126 to 138 (Quantile 80 - 90%)',
'Predictive Feature: Votes\nFeature Value: 61 to 4870.8 (Quantile 0 - 10%)',
'Predictive Feature: Year\nFeature Value: Higher than 2016 (Quantile 100%)',
]


IMDB_X_NUM_FIRST_ROW = [
0.0,
0.0,
0.0,
0.0,
0.0,
0.0,
-0.10520633310079575,
0.5200179815292358,
-3.0,
0.0,
1.0936769247055054,
-0.9359688758850098,
0.9988187551498413,
]


IMDB_X_TXT_LAST_ROW = [
'Target Feature: Genre\nFeature Value: Drama',
'Target Feature: Genre\nFeature Value: Not Drama',
'Predictive Feature: Title\nFeature Value: Hunger',
'Predictive Feature: Description\nFeature Value: Irish republican Bobby Sands leads the inmates of a Northern Irish prison in a hunger strike.',
'Predictive Feature: Director\nFeature Value: Steve McQueen',
'Predictive Feature: Actors\nFeature Value: Stuart Graham, Laine Megaw, Brian Milligan, Liam McMahon',
'Predictive Feature: Metascore\nFeature Value: 81 to 100 (Quantile 90 - 100%)',
'Predictive Feature: Rank\nFeature Value: 598.4 to 691.3 (Quantile 60 - 70%)',
'Predictive Feature: Rating\nFeature Value: 7.5 to 7.9 (Quantile 80 - 90%)',
'Predictive Feature: Revenue (Millions)\nFeature Value: 0 to 1.816 (Quantile 0 - 10%)',
'Predictive Feature: Runtime (Minutes)\nFeature Value: 91 to 97 (Quantile 10 - 20%)',
'Predictive Feature: Votes\nFeature Value: 50250.8 to 78228.4 (Quantile 30 - 40%)',
'Predictive Feature: Year\nFeature Value: 2008 to 2010 (Quantile 10 - 20%)',
]

IMDB_X_NUM_LAST_ROW = [
0.0,
0.0,
0.0,
0.0,
0.0,
0.0,
1.3703500032424927,
0.6548995971679688,
0.9240633249282837,
-0.8019106984138489,
-0.9082886576652527,
-0.6306809782981873,
-1.502341866493225,
]


# def test_tabstar_preprocessing_imdb(): # this test breaks and its fix is not prioritized for now
#     _rm_temp_dir()
#     data_dir = create_pretrain_dataset(dataset_id=OpenMLDatasetID.BIN_SOCIAL_IMDB_GENRE_PREDICTION,
#                                        cache_dir=TEST_DATA_DIR)
#     dev_dataloader = get_dev_dataloader(data_dir=data_dir, batch_size=40)
#     x_txt, x_num, y, properties = next(iter(dev_dataloader))
#     assert properties.d_output == 2
#     assert properties.train_size == 760
#     assert properties.val_size == 40
#     assert np.mean(y.numpy()) == 0.475

#     # First example
#     assert y[0] == 1
#     assert x_txt[0].tolist() == IMDB_X_TXT_FIRST_ROW
#     assert x_num[0].tolist() == IMDB_X_NUM_FIRST_ROW

#     # Last example
#     assert y[-1] == 0
#     assert x_txt[-1].tolist() == IMDB_X_TXT_LAST_ROW
#     assert x_num[-1].tolist() == IMDB_X_NUM_LAST_ROW
#     _rm_temp_dir()

def _rm_temp_dir():
    if os.path.isdir(TEST_DATA_DIR):
        shutil.rmtree(TEST_DATA_DIR)


import torch
from torch import nn, Tensor

from tabstar.arch.config import D_MODEL


class NumericalFusion(nn.Module):

    def __init__(self):
        super().__init__()
        self.scalar_embedder = nn.Sequential(
            nn.Linear(1, D_MODEL * 2),
            nn.ReLU(),
            nn.Dropout(0.1),
            nn.Linear(D_MODEL * 2, D_MODEL)
        )
        self.fusion_block = nn.TransformerEncoderLayer(
            d_model=D_MODEL,
            nhead=2,
            dim_feedforward=D_MODEL * 4,
            dropout=0.1,
            activation='relu',
            batch_first=True,
            norm_first=True
        )


    def forward(self, textual_embeddings: Tensor, x_num: Tensor) -> Tensor:
        batch_size, seq_len, d_model = textual_embeddings.shape
        num_embeddings = self.scalar_embedder(x_num.unsqueeze(-1))
        assert num_embeddings.shape == textual_embeddings.shape
        fusion_input = torch.stack([textual_embeddings, num_embeddings], dim=2)
        assert fusion_input.shape == (batch_size, seq_len, 2, d_model)
        fusion_input = fusion_input.view(batch_size * seq_len, 2, d_model)
        fused = self.fusion_block(fusion_input)
        fused_embeddings = fused.view(batch_size, seq_len, 2, d_model)
        fused_embeddings = fused_embeddings.mean(dim=2)
        assert fused_embeddings.shape == textual_embeddings.shape
        return fused_embeddings



from tabstar.preprocessing.texts import replace_whitespaces
from tabstar.preprocessing.verbalize import verbalize_feature


def test_replace_whitespace():
    assert replace_whitespaces('a text\nwith newline') == 'a text with newline'
    assert replace_whitespaces('a text   with multiple spaces') == 'a text   with multiple spaces'


def test_verbalize_feature():
    f = verbalize_feature(col="f1", value="value1")
    assert f == "Predictive Feature: f1\nFeature Value: value1"


from torch import nn
from torch.optim import AdamW
from torch.optim.lr_scheduler import OneCycleLR, LRScheduler

WARMUP_PROPORTION = 0.1


def get_optimizer(model: nn.Module, lr: float) -> AdamW:
    params = [{"params": model.parameters(), "lr": lr, "name": "lora_lr"}]
    optimizer = AdamW(params)
    return optimizer

def get_scheduler(optimizer: AdamW, max_lr: float, epochs: int) -> LRScheduler:
    # TODO: we currently use total_steps=epochs, but it expects to be affected by accumulation steps.
    # We might want to reconsider setting this depending on the expected number of updates.
    try:
        return OneCycleLR(optimizer=optimizer, max_lr=max_lr, total_steps=epochs,
                          pct_start=WARMUP_PROPORTION, anneal_strategy='cos')
    except ZeroDivisionError:
        return OneCycleLR(optimizer=optimizer, max_lr=max_lr, total_steps=epochs+1,
                          pct_start=WARMUP_PROPORTION, anneal_strategy='cos')


from typing import Tuple, Optional

from pandas import DataFrame, Series, SparseDtype


def densify_objects(x: DataFrame, y: Optional[Series]) -> Tuple[DataFrame, Optional[Series]]:
    for col in x.columns:
        x[col] = densify_series(x[col])
    if y is not None:
        y = densify_series(y)
    return x, y



def densify_series(s: Series) -> Series:
    if not isinstance(s.dtype, SparseDtype):
        return s
    return s.sparse.to_dense()